import os
import json
import math
from typing import Union, Tuple, List

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from easy_mpl import plot
from easy_mpl import bar_chart, taylor_plot, dumbbell_plot
from SeqMetrics import RegressionMetrics, ClassificationMetrics

from ai4water.backend import tf
from ai4water.hyperopt import HyperOpt
from ai4water.preprocessing import DataSet
from ai4water.utils.utils import jsonize, ERROR_LABELS
from ai4water.postprocessing import ProcessResults 
from ai4water.utils.utils import clear_weights, dateandtime_now, dict_to_file


if tf is not None:
    if 230 <= int(''.join(tf.__version__.split('.')[0:2]).ljust(3, '0')) < 250:
        from ai4water.functional import Model
        print(f"""
        Switching to functional API due to tensorflow version {tf.__version__} 
        for experiments""")
    else:
        from ai4water import Model
else:
    from ai4water import Model

SEP = os.sep

# todo plots comparing different models in following youtube videos at 6:30 and 8:00 minutes.
# https://www.youtube.com/watch?v=QrJlj0VCHys
# compare models using statistical tests wuch as Giacomini-White test or Diebold-Mariano test
# paired ttest 5x2cv


# in order to unify the use of metrics
Metrics = {
    'regression': lambda t, p, multiclass=False, **kwargs: RegressionMetrics(t, p, **kwargs),
    'classification': lambda t, p, multiclass=False, **kwargs: ClassificationMetrics(t, p, 
        multiclass=multiclass, **kwargs)
}

Monitor = {
    'regression': ['r2', 'corr_coeff', 'mse', 'rmse', 'r2_score',
                'nse', 'kge', 'mape', 'pbias', 'bias', 'mae', 'nrmse',
                'mase'],

    'classification': ['accuracy', 'precision', 'recall', 'mse']
}

class Experiments(object):
    """
    Base class for all the experiments.

    All the expriments must be subclasses of this class.
    The core idea of ``Experiments`` is based upon ``model``. An experiment 
    consists of one or more models. The models differ from each other in their 
    structure/idea/concept/configuration. When :py:meth:`ai4water.experiments.Experiments.fit` 
    is called, each ``model`` is built and trained. The user can customize, building 
    and training process by subclassing this class and customizing 
    :py:meth:`ai4water.experiments.Experiments._build` and 
    :py:meth:`ai4water.experiments.Experiments._fit` methods.

    Attributes
    ------------
    - metrics
    - exp_path
    - model_
    - models

    Methods
    --------
    - fit
    - taylor_plot
    - compare_losses
    - plot_convergence
    - from_config

    """
    def __init__(
            self,
            cases: dict = None,
            exp_name: str = None,
            num_samples: int = 5,
            verbosity: int = 1,
            monitor: Union[str, list] = None,
    ):
        """

        Arguments
        ---------
            cases :
                python dictionary defining different cases/scenarios
            exp_name :
                name of experiment, used to define path in which results are saved
            num_samples :
                only relevent when you wan to optimize hyperparameters of models
                using ``grid`` method
            verbosity : bool, optional
                determines the amount of information
            monitor : str, list, optional
                list of performance metrics to monitor. It can be any performance
                metric `SeqMetrics_ <https://seqmetrics.readthedocs.io>` library.
                By default ``r2``, ``corr_coeff, ``mse``, ``rmse``, ``r2_score``,
                ``nse``, ``kge``, ``mape``, ``pbias``, ``bias`` ``mae``, ``nrmse``
                ``mase`` are considered for regression and ``accuracy``, ``precision``
                ``recall`` are considered for classification.
        """
        self.opt_results = None
        self.optimizer = None
        self.exp_name = 'Experiments_' + str(dateandtime_now()) if exp_name is None else exp_name
        self.num_samples = num_samples
        self.verbosity = verbosity

        self.models = [method for method in dir(self) if callable(getattr(self, method)) if method.startswith('model_')]

        if cases is None:
            cases = {}
        self.cases = {'model_'+key if not key.startswith('model_') else key: val for key, val in cases.items()}
        self.models = self.models + list(self.cases.keys())

        self.exp_path = os.path.join(os.getcwd(), "results", self.exp_name)
        if not os.path.exists(self.exp_path):
            os.makedirs(self.exp_path)

        self.update_config(models=self.models, exp_path=self.exp_path,
            exp_name=self.exp_name, cases=self.cases)

        if monitor is None:
            self.monitor = Monitor[self.mode]

    def update_and_save_config(self, **kwargs):
        self.update_config(**kwargs)
        self.save_config()
        return

    def update_config(self, **kwargs):
        if not hasattr(self, 'config'):
            setattr(self, 'config', {})
        self.config.update(kwargs.copy())
        return

    def save_config(self):
        dict_to_file(self.exp_path, config=self.config)
        return

    def build_from_config(self, data, config_path, weights, **kwargs):
        setattr(self, 'model_', None)
        raise NotImplementedError

    @property
    def tpot_estimator(self):
        raise NotImplementedError

    @property
    def mode(self):
        raise NotImplementedError

    @property
    def num_samples(self):
        return self._num_samples

    @num_samples.setter
    def num_samples(self, x):
        self._num_samples = x

    def _reset(self):

        self.cv_scores_ = {}

        self.config['eval_models'] = {}
        self.config['optimized_models'] = {}

        self.metrics = {}
        self.features = {}
        self.iter_metrics = {}

        return

    def fit(
            self,
            data,
            run_type: str = "dry_run",
            opt_method: str = "bayes",
            num_iterations: int = 12,
            include: Union[None, list] = None,
            exclude: Union[None, list, str] = '',
            cross_validate: bool = False,
            post_optimize: str = 'eval_best',
            hpo_kws: dict = None
    ):
        """
        Runs the fit loop for all the ``models`` of experiment. The user can
        however, specify the models by making use of ``include`` and ``exclud``
        keywords.

        todo, post_optimize not working for 'eval_best' with ML methods.

        Arguments:
            data: this will be passed to :py:meth:`ai4water.Model.fit`.
            run_type :
                One of ``dry_run`` or ``optimize``. If ``dry_run``, the all
                the `models` will be trained only once. if ``optimize``, then
                hyperparameters of all the models will be optimized.
            opt_method :
                which optimization method to use. options are ``bayes``,
                ``random``, ``grid``. Only valid if ``run_type`` is ``optimize``
            num_iterations : number of iterations for optimization. Only valid
                if ``run_type`` is ``optimize``.
            include :
                name of models to included. If None, all the models found
                will be trained and or optimized.
            exclude :
                name of ``models`` to be excluded
            cross_validate :
                whether to cross validate the model or not. This
                depends upon `cross_validator` agrument to the `Model`.
            post_optimize :
                one of ``eval_best`` or ``train_best``. If eval_best,
                the weights from the best models will be uploaded again and the model
                will be evaluated on train, test and all the data. If ``train_best``,
                then a new model will be built and trained using the parameters of
                the best model.
            hpo_kws :
                keyword arguments for :py:class:`ai4water.hyperopt.HyperOpt` class.

        Examples
        ---------
        >>> from ai4water.experiments import MLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> exp = MLRegressionExperiments()
        >>> exp.fit(data=busan_beach())

        If you want to compare only RandomForest, XGBRegressor, CatBoostRegressor
        and LGBMRegressor, use the ``include`` keyword

        >>> exp.fit(data=busan_beach(), include=['RandomForestRegressor', 'XGBRegressor',
        >>>    'CatBoostRegressor', 'LGBMRegressor'])

        Similarly, if you want to exclude certain models from comparison, you can
        use ``exclude`` keyword

        >>> exp.fit(data=busan_beach(), exclude=["SGDRegressor"])

        if you want to perform cross validation for each model, we must give
        the ``cross_validator`` argument which will be passed to ai4water Model

        >>> exp = MLRegressionExperiments(cross_validator={"KFold": {"n_splits": 10}})
        >>> exp.fit(data=busan_beach(), cross_validate=True)

        Setting ``cross_validate`` to True will populate `cv_scores_` dictionary
        which can be accessed as ``exp.cv_scores_``

        if you want to optimize the hyperparameters of each model,

        >>> exp.fit(data=busan_beach(), run_type="optimize", num_iterations=20)

        """
        assert run_type in ['optimize', 'dry_run']

        assert post_optimize in ['eval_best', 'train_best']

        if exclude == '':
            exclude = []

        predict = False
        if run_type == 'dry_run':
            predict = True

        if hpo_kws is None:
            hpo_kws = {}

        include = self._check_include_arg(include)

        if exclude is None:
            exclude = []
        elif isinstance(exclude, str):
            exclude = [exclude]

        consider_exclude(exclude, self.models, include)

        self._reset()

        for model_type in include:

            model_name = model_type.split('model_')[1]

            if model_type not in exclude:

                self.model_iter_metric = {}
                self.iter_ = 0

                def objective_fn(**suggested_paras):
                    # the config must contain the suggested parameters by the hpo algorithm
                    if model_type in self.cases:
                        config = self.cases[model_type]
                        config.update(suggested_paras)
                    elif model_name in self.cases:
                        config = self.cases[model_name]
                        config.update(suggested_paras)
                    elif hasattr(self, model_type):
                        config = getattr(self, model_type)(**suggested_paras)
                    else:
                        raise TypeError

                    return self._build_and_run(
                        data=data,
                        predict=predict,
                        cross_validate=cross_validate,
                        title=f"{self.exp_name}{SEP}{model_name}",
                        **config)

                if run_type == 'dry_run':
                    if self.verbosity >= 0: print(f"running  {model_type} model")
                    train_results, test_results = objective_fn()
                    self._populate_results(model_name, train_results, test_results)
                else:
                    # there may be attributes int the model, which needs to be loaded so run the method first.
                    # such as param_space etc.
                    if hasattr(self, model_type):
                        getattr(self, model_type)()

                    opt_dir = os.path.join(os.getcwd(),
                        f"results{SEP}{self.exp_name}{SEP}{model_name}")

                    if self.verbosity > 0: 
                        print(f"optimizing  {model_type} using {opt_method} method")

                    self.optimizer = HyperOpt(
                        opt_method,
                        objective_fn=objective_fn,
                        param_space=self.param_space, 
                        opt_path=opt_dir,
                        num_iterations=num_iterations,  # number of iterations
                        x0=self.x0,
                        verbosity=self.verbosity,
                        **hpo_kws
                        )

                    self.opt_results = self.optimizer.fit()

                    self.config['optimized_models'][model_type] = self.optimizer.opt_path

                    if post_optimize == 'eval_best':
                        self.eval_best(data, model_name, opt_dir)
                    elif post_optimize == 'train_best':
                        self.train_best(data, model_name)

                if not hasattr(self, 'model_'):  # todo asking user to define this parameter is not good
                    raise ValueError(f'The `build` method must set a class level attribute named `model_`.')
                self.config['eval_models'][model_type] = self.model_.path

                if cross_validate:
                    cv_scoring = self.model_.val_metric
                    self.cv_scores_[model_type] = getattr(self.model_, f'cross_val_{cv_scoring}')
                    setattr(self, '_cv_scoring', cv_scoring)

                self.iter_metrics[model_type] = self.model_iter_metric

        self.save_config()

        save_json_file(os.path.join(self.exp_path, 'features.json'), self.features)
        save_json_file(os.path.join(self.exp_path, 'metrics.json'), self.metrics)

        return

    def eval_best(self, data, model_type, opt_dir, **kwargs):
        """Evaluate the best models."""
        best_models = clear_weights(opt_dir, rename=False, write=False)
        # TODO for ML, best_models is empty
        if len(best_models) < 1:
            return self.train_best(data, model_type)
        for mod, props in best_models.items():
            mod_path = os.path.join(props['path'], "config.json")
            mod_weights = props['weights']

            train_results, test_results = self.build_from_config(
                data,
                mod_path, 
                mod_weights, 
                **kwargs)

            if mod.startswith('1_'):
                self._populate_results(model_type, train_results, test_results)
        return

    def train_best(self, data, model_type):
        """Train the best model."""
        best_paras = self.optimizer.best_paras()
        if best_paras.get('lookback', 1) > 1:
            _model = 'layers'
        else:
            _model = model_type
        train_results, test_results = self._build_and_run(
            data=data,
            predict=True,
            model={_model: self.optimizer.best_paras()},
            title=f"{self.exp_name}{SEP}{model_type}{SEP}best"
            )

        self._populate_results(model_type, train_results, test_results)
        return

    def _populate_results(self, model_type: str,
                          train_results: Tuple[np.ndarray, np.ndarray],
                          test_results: Tuple[np.ndarray, np.ndarray]
                          ):

        if not model_type.startswith('model_'):  # internally we always use model_ at the start.
            model_type = f'model_{model_type}'

        # save standard deviation on true and predicted arrays of train and test
        self.features[model_type] = {
            'train': {
                'true': {
                    'std': np.std(train_results[0])},
                'simulation':{
                    'std': np.std(train_results[1])}},

            'test': {
                'true': {
                    'std': np.std(test_results[0])},
                'simulation': {
                    'std': np.std(test_results[1])}}}

        # save performance metrics of train and test
        metrics = Metrics[self.mode](train_results[0],
                                     train_results[1],
                                     multiclass=self.model_.is_multiclass)
        train_metrics = {}
        for metric in self.monitor:
            train_metrics[metric] = getattr(metrics, metric)()

        metrics = Metrics[self.mode](test_results[0],
                                     test_results[1],
                                     multiclass=self.model_.is_multiclass)
        test_metrics = {}
        for metric in self.monitor:
            test_metrics[metric] = getattr(metrics, metric)()

        self.metrics[model_type] = {'train': train_metrics,
                                    'test': test_metrics}
        return

    def taylor_plot(
            self,
            include: Union[None, list] = None,
            exclude: Union[None, list] = None,
            figsize: tuple = (9, 7),
            **kwargs
    )->plt.Figure:
        """
        Compares the models using taylor_plot_.

        Parameters
        ----------
            include : str, list, optional
                if not None, must be a list of models which will be included.
                None will result in plotting all the models.
            exclude : str, list, optional
                if not None, must be a list of models which will excluded.
                None will result in no exclusion
            figsize : tuple, optional
            **kwargs :
                all the keyword arguments for taylor_plot_ function.

        Returns
        -------
        plt.Figure

        Example
        -------
        >>> from ai4water.experiments import MLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> data = busan_beach()
        >>> inputs = list(data.columns)[0:-1]
        >>> outputs = list(data.columns)[-1]
        >>> experiment = MLRegressionExperiments(input_features=inputs, output_features=outputs)
        >>> experiment.fit(data=data)
        >>> experiment.taylor_plot()

        .. _taylor_plot:
            https://easy-mpl.readthedocs.io/en/latest/#easy_mpl.taylor_plot
        """
        metrics = self.metrics.copy()

        include = self._check_include_arg(include)

        if exclude is not None:
            consider_exclude(exclude, self.models, metrics)

        if 'name' in kwargs:
            fname = kwargs.pop('name')
        else:
            fname = 'taylor'
        fname = os.path.join(os.getcwd(), f'results{SEP}{self.exp_name}{SEP}{fname}.png')

        train_std = [_model['train']['true']['std'] for _model in self.features.values()]
        train_std = list(set(train_std))[0]
        test_std = [_model['test']['true']['std'] for _model in self.features.values()]
        test_std = list(set(test_std))[0]

        observations = {'train': {'std': train_std},
                        'test': {'std': test_std}}

        simulations = {'train': None, 'test': None}

        for scen in ['train', 'test']:

            scen_stats = {}
            for model, _metrics in metrics.items():
                model_stats = {'std': self.features[model][scen]['simulation']['std'],
                               'corr_coeff': _metrics[scen]['corr_coeff'],
                               'pbias': _metrics[scen]['pbias']
                               }

                if model in include:
                    key = shred_model_name(model)
                    scen_stats[key] = model_stats

            simulations[scen] = scen_stats

        return taylor_plot(
            observations=observations,
            simulations=simulations,
            figsize=figsize,
            name=fname,
            **kwargs
        )

    def _consider_include(self, include: Union[str, list], to_filter):

        filtered = {}

        include = self._check_include_arg(include)

        for m in include:
            if m in to_filter:
                filtered[m] = to_filter[m]

        return filtered

    def _check_include_arg(self, include):

        if isinstance(include, str):
            include = [include]

        if include is None:
            include = self.models
        include = ['model_' + _model if not _model.startswith('model_') else _model for _model in include]

        # make sure that include contains same elements which are present in models
        for elem in include:
            assert elem in self.models, f"""
{elem} to `include` are not available.
Available cases are {self.models} and you wanted to include
{include}
"""
        return include

    def plot_improvement(
            self,
            metric_name: str,
            plot_type: str = 'dumbell',
            save: bool = True,
            name : str = '',
            dpi : int = 200,
            **kwargs
    ) -> pd.DataFrame:
        """
        Shows how much improvement was observed after hyperparameter
        optimization. This plot is only available if ``run_type`` was set to
        `optimize` in :py:meth:`ai4water.experiments.Experiments.fit`.

        Arguments
        ---------
            metric_name :
                the peformance metric for comparison
            plot_type : str, optional
                the kind of plot to draw. Either ``dumbell`` or ``bar``
            save : bool
                whether to save the plot or not
            name : str, optional
            dpi : int, optional
            **kwargs :
                any additional keyword arguments for
                `dumbell plot <https://easy-mpl.readthedocs.io/en/latest/#easy_mpl.dumbbell_plot>`_
                 or `bar_chart <https://easy-mpl.readthedocs.io/en/latest/#easy_mpl.bar_chart>`_

        Returns
        -------
        pd.DataFrame

        Examples
        --------
        >>> from ai4water.experiments import MLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> experiment = MLRegressionExperiments()
        >>> experiment.fit(data=busan_beach(), run_type="optimize", num_iterations=30)
        >>> experiment.plot_improvement('r2')

        or draw dumbell plot

        >>> experiment.plot_improvement('r2', plot_type='bar')

        """

        data: str = 'test'

        assert data in ['training', 'test', 'validation']

        improvement = pd.DataFrame(columns=['start', 'end'])

        for model, model_iter_metrics in self.iter_metrics.items():
            initial = model_iter_metrics[0][metric_name]
            final = self.metrics[model]['test'][metric_name]

            key = shred_model_name(model)

            improvement.loc[key] = [initial, final]

        if plot_type == "dumbell":
            ax = dumbbell_plot(
                improvement['start'], improvement['end'],
                improvement.index.tolist(),
                xlabel=ERROR_LABELS.get(metric_name, metric_name),
                show=False,
                **kwargs
            )
            ax.set_xlabel(ERROR_LABELS.get(metric_name, metric_name))
        else:

            colors = {
                'start': np.array([0, 56, 104]) / 256,
                'end': np.array([126, 154, 178]) / 256
            }

            order = ['start', 'end']
            if metric_name in ['r2', 'nse', 'kge', 'corr_coeff', 'r2_mod', 'r2_score']:
                order = ['end', 'start']

            fig, ax = plt.subplots()

            for ordr in order:
                bar_chart(improvement[ordr], improvement.index.tolist(),
                          ax=ax, color=colors[ordr], show=False,
                          xlabel=ERROR_LABELS.get(metric_name, metric_name),
                          label=ordr, **kwargs)

            ax.legend()
            plt.title('Improvement after Optimization')

        if save:
            fname = os.path.join(
                os.getcwd(),
                f'results{SEP}{self.exp_name}{SEP}{name}_improvement_{metric_name}.png')
            plt.savefig(fname, dpi=dpi, bbox_inches=kwargs.get('bbox_inches', 'tight'))

        plt.show()

        return improvement

    def compare_errors(
            self,
            matric_name: str,
            cutoff_val: float = None,
            cutoff_type: str = None,
            save: bool = True,
            sort_by: str = 'test',
            ignore_nans: bool = True,
            name: str = 'ErrorComparison',
            show: bool = True,
            **kwargs
    ) -> pd.DataFrame:
        """
        Plots a specific performance matric for all the models which were
        run during [fit][ai4water.experiments.Experiments.fit] call.

        Parameters
        ----------
            matric_name:
                 performance matric whose value to plot for all the models
            cutoff_val:
                 if provided, only those models will be plotted for whome the
                 matric is greater/smaller than this value. This works in conjuction
                 with `cutoff_type`.
            cutoff_type:
                 one of ``greater``, ``greater_equal``, ``less`` or ``less_equal``.
                 Criteria to determine cutoff_val. For example if we want to
                 show only those models whose $R^2$ is > 0.5, it will be 'max'.
            save:
                whether to save the plot or not
            sort_by:
                either ``test`` or ``train``. How to sort the results for plotting.
                If 'test', then test performance matrics will be sorted otherwise
                train performance matrics will be sorted.
            ignore_nans:
                default True, if True, then performance matrics with nans are ignored
                otherwise nans/empty bars will be shown to depict which models have
                resulted in nans for the given performance matric.
            name:
                name of the saved file.
            show : whether to show the plot at the end or not?

            kwargs :

                - fig_height :
                - fig_width :
                - title_fs :
                - xlabel_fs :
                - color :

        returns
        -------
        pd.DataFrame
            pandas dataframe whose index is models and has two columns with name
            'train' and 'test' These columns contain performance metrics for the
            models..

        Example
        -------
            >>> from ai4water.experiments import MLRegressionExperiments
            >>> from ai4water.datasets import busan_beach
            >>> data = busan_beach()
            >>> inputs = list(data.columns)[0:-1]
            >>> outputs = list(data.columns)[-1]
            >>> experiment = MLRegressionExperiments(input_features=inputs, output_features=outputs)
            >>> experiment.fit(data=data)
            >>> experiment.compare_errors('mse')
            >>> experiment.compare_errors('r2', 0.2, 'greater')
        """

        models = self.sort_models_by_metric(matric_name, cutoff_val, cutoff_type,
                                            ignore_nans, sort_by)

        plt.close('all')
        fig, axis = plt.subplots(1, 2, sharey='all')
        fig.set_figheight(kwargs.get('fig_height', 8))
        fig.set_figwidth(kwargs.get('fig_width', 8))

        bar_chart(ax=axis[0],
                  labels=models.index.tolist(),
                  values=models['train'],
                  color=kwargs.get('color', None),
                  title="Train",
                  xlabel=ERROR_LABELS.get(matric_name, matric_name),
                  xlabel_fs=kwargs.get('xlabel_fs', 16),
                  title_fs=kwargs.get('title_fs', 20),
                  show=False,
                  )

        bar_chart(ax=axis[1],
                  labels=models.index.tolist(),
                  values=models['test'],
                  title="Test",
                  color=kwargs.get('color', None),
                  xlabel=ERROR_LABELS.get(matric_name, matric_name),
                  xlabel_fs=kwargs.get('xlabel_fs', 16),
                  title_fs=kwargs.get('title_fs', 20),
                  show_yaxis=False,
                  show=False
                  )

        appendix = f"{cutoff_val or ''}{cutoff_type or ''}{len(models)}"
        if save:
            fname = os.path.join(
                os.getcwd(),
                f'results{SEP}{self.exp_name}{SEP}{name}_{matric_name}_{appendix}.png')
            plt.savefig(fname, dpi=100, bbox_inches=kwargs.get('bbox_inches', 'tight'))

        if show:
            plt.show()
        return models

    def loss_comparison(
            self,
            loss_name: Union[str, list] = 'loss',
            include: list = None,
            save: bool = True,
            show: bool = True,
            figsize: int = None,
            start: int = 0,
            end: int = None,
            **kwargs
    ) -> plt.Axes:
        """
        Plots the loss curves of the evaluated models. This method is only available
        if the models which are being compared are deep leanring mdoels.

        Parameters
        ----------
            loss_name : str, optional
                the name of loss value, must be recorded during training
            include:
                name of models to include
            save:
                whether to save the plot or not
            name:
                name of saved file
            show:
                whether to show the plot or now
            figsize : tuple
                size of the figure
            **kwargs :
                any other keyword arguments to be passed to the 
                `plot <https://easy-mpl.readthedocs.io/en/latest/#easy_mpl.plot>`_
        Returns:
            matplotlib axes

        Example
        -------
        >>> from ai4water.experiments import DLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> data = busan_beach()
        >>> exp = DLExperiments(
        >>> input_features = data.columns.tolist()[0:-1],
        >>> output_features = data.columns.tolist()[-1:],
        >>> epochs=300,
        >>> train_fraction=1.0,
        >>> y_transformation="log",
        >>> x_transformation="minmax",
        >>> )

        >>> exp.fit(data=data)
        >>> exp.loss_comparison()

        you may wish to plot on log scale

        >>> exp.loss_comparison(logy=True)
        """

        if not isinstance(loss_name, list):
            assert isinstance(loss_name, str)
            loss_name = [loss_name]

        include = self._check_include_arg(include)

        loss_curves = {}
        for _model, _path in self.config['eval_models'].items():
            if _model in include:
                df = pd.read_csv(os.path.join(_path, 'losses.csv'), usecols=loss_name)
                loss_curves[_model] = df.values

        _kwargs = {'linestyle': '-',
                   'xlabel': "Epochs",
                   'ylabel': 'Loss'}

        if len(loss_curves) > 5:
            _kwargs['legend_kws'] = {'bbox_to_anchor': (1.1, 0.99)}

        end = end or len(df)

        _, axis = plt.subplots(figsize=figsize)

        for _model, _loss in loss_curves.items():
            label = shred_model_name(_model)
            plot(_loss[start:end], ax=axis, label=label, show=False, **_kwargs, **kwargs)

        if save:
            fname = os.path.join(self.exp_path, f'loss_comparison_{loss_name}.png')
            plt.savefig(fname, dpi=100, bbox_inches='tight')

        if show:
            plt.show()

        return axis

    def compare_convergence(
            self,
            show: bool = True,
            save: bool = False,
            name: str = 'convergence_comparison',
            **kwargs
    ) -> Union[plt.Axes, None]:
        """
        Plots and compares the convergence plots of hyperparameter optimization runs.
        Only valid if `run_type=optimize` during :py:meth:`ai4water.experiments.Experiments.fit`
        call.

        Parameters
        ----------
            show :
                whether to show the plot or now
            save :
                whether to save the plot or not
            name :
                name of file to save the plot
            kwargs :
                keyword arguments to plot_ function
        Returns
        -------
            if the optimized models are >1 then it returns the maplotlib axes
            on which the figure is drawn otherwise it returns None.

        Examples
        --------
        >>> from ai4water.experiments import MLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> experiment = MLRegressionExperiments()
        >>> experiment.fit(data=busan_beach(), run_type="optimize", num_iterations=30)
        >>> experiment.compare_convergence()

        .. _plot:
            https://easy-mpl.readthedocs.io/en/latest/#easy_mpl.plot
        """
        if len(self.config['optimized_models']) < 1:
            print('No model was optimized')
            return
        plt.close('all')
        fig, axis = plt.subplots()

        for _model, opt_path in self.config['optimized_models'].items():
            with open(os.path.join(opt_path, 'iterations.json'), 'r') as fp:
                iterations = json.load(fp)

            convergence = sort_array(list(iterations.keys()))

            label = shred_model_name(_model)
            plot(
                convergence,
                ax=axis,
                label=label,
                linestyle='--',
                xlabel='Number of iterations $n$',
                ylabel=r"$\min f(x)$ after $n$ calls",
                show=False,
                **kwargs
            )
        if save:
            fname = os.path.join(self.exp_path, f'{name}.png')
            plt.savefig(fname, dpi=100, bbox_inches='tight')

        if show:
            plt.show()
        return axis

    @classmethod
    def from_config(
            cls,
            config_path: str,
            **kwargs
    ) -> "Experiments":
        """
        Loads the experiment from the config file.

        Arguments:
            config_path : complete path of experiment
            kwargs : keyword arguments to experiment
        Returns:
            an instance of `Experiments` class
        """
        if not config_path.endswith('.json'):
            raise ValueError(f"""
{config_path} is not a json file
""")
        with open(config_path, 'r') as fp:
            config = json.load(fp)

        cls.config = config
        cls._from_config = True

        cv_scores = {}
        scoring = "mse"

        for model_name, model_path in config['eval_models'].items():

            with open(os.path.join(model_path, 'config.json'), 'r') as fp:
                model_config = json.load(fp)

            # if cross validation was performed, then read those results.
            cross_validator = model_config['config']['cross_validator']
            if cross_validator is not None:
                cv_name = str(list(cross_validator.keys())[0])
                scoring = model_config['config']['val_metric']
                cv_fname = os.path.join(model_path, f'{cv_name}_{scoring}' + ".json")
                if os.path.exists(cv_fname):
                    with open(cv_fname, 'r') as fp:
                        cv_scores[model_name] = json.load(fp)

        cls.metrics = load_json_file(
            os.path.join(os.path.dirname(config_path), "metrics.json"))
        cls.features = load_json_file(
            os.path.join(os.path.dirname(config_path), "features.json"))
        cls.cv_scores_ = cv_scores
        cls._cv_scoring = scoring

        return cls(exp_name=config['exp_name'], cases=config['cases'], **kwargs)

    def plot_cv_scores(
            self,
            show: bool = False,
            name: str = "cv_scores",
            exclude: Union[str, list] = None,
            include: Union[str, list] = None,
            **kwargs
    ) -> Union[plt.Axes, None]:
        """
        Plots the box whisker plots of the cross validation scores.

        This plot is only available if cross_validation was set to True during
        :py:meth:`ai4water.experiments.Experiments.fit`.

        Arguments
        ---------
            show : whether to show the plot or not
            name : name of the plot
            include : models to include
            exclude : models to exclude
            kwargs : any of the following keyword arguments

                - notch
                - vert
                - figsize
                - bbox_inches

        Returns
        -------
            matplotlib axes if the figure is drawn otherwise None

        Example
        -------
        >>> from ai4water.experiments import MLRegressionExperiments
        >>> from ai4water.datasets import busan_beach
        >>> exp = MLRegressionExperiments(cross_validator={"KFold": {"n_splits": 10}})
        >>> exp.fit(data=busan_beach(), cross_validate=True)
        >>> exp.plot_cv_scores()

        """
        if len(self.cv_scores_) == 0:
            return

        scoring = self._cv_scoring
        cv_scores = self.cv_scores_

        consider_exclude(exclude, self.models, cv_scores)
        cv_scores = self._consider_include(include, cv_scores)

        model_names = [m.split('model_')[1] for m in list(cv_scores.keys())]
        if len(model_names) < 5:
            rotation = 0
        else:
            rotation = 90

        plt.close()

        _, axis = plt.subplots(figsize=kwargs.get('figsize', (8, 6)))

        d = axis.boxplot(list(cv_scores.values()),
                         notch=kwargs.get('notch', None),
                         vert=kwargs.get('vert', None),
                         labels=model_names
                         )

        whiskers = d['whiskers']
        caps = d['caps']
        boxes = d['boxes']
        medians = d['medians']
        fliers = d['fliers']

        axis.set_xticklabels(model_names, rotation=rotation)
        axis.set_xlabel("Models", fontsize=16)
        axis.set_ylabel(ERROR_LABELS.get(scoring, scoring), fontsize=16)

        fname = os.path.join(os.getcwd(),
                             f'results{SEP}{self.exp_name}{SEP}{name}_{len(model_names)}.png')
        plt.savefig(fname, dpi=300, bbox_inches=kwargs.get('bbox_inches', 'tight'))

        if show:
            plt.show()

        return axis

    def sort_models_by_metric(
            self,
            metric_name,
            cutoff_val=None,
            cutoff_type=None,
            ignore_nans: bool = True,
            sort_by="test"
    ) -> pd.DataFrame:
        """returns the models sorted according to their performance"""

        idx = list(self.metrics.keys())
        train_metrics = [v['train'][metric_name] for v in self.metrics.values()]
        test_metrics = [v['test'][metric_name] for v in self.metrics.values()]

        df = pd.DataFrame(np.vstack((train_metrics, test_metrics)).T,
                          index=idx,
                          columns=['train', 'test'])
        if ignore_nans:
            df = df.dropna()

        df = df.sort_values(by=[sort_by], ascending=False)

        if cutoff_type is not None:
            assert cutoff_val is not None
            if cutoff_type=="greater":
                df = df.loc[df[sort_by]>cutoff_val]
            else:
                df = df.loc[df[sort_by]<cutoff_val]

        return df

    def fit_with_tpot(
            self,
            data,
            models: Union[int, List[str], dict, str] = None,
            selection_criteria: str = 'mse',
            scoring: str = None,
            **tpot_args
    ):
        """
        Fits the tpot_'s fit  method which
        finds out the best pipline for the given data.

        Arguments
        ---------
            data :
            models :
                It can be of three types.

                - If list, it will be the names of machine learning models/
                    algorithms to consider.
                - If integer, it will be the number of top
                    algorithms to consider for tpot. In such a case, you must have
                    first run `.fit` method before running this method. If you run
                    the tpot using all available models, it will take hours to days
                    for medium sized data (consisting of few thousand examples). However,
                    if you run first .fit and see for example what are the top 5 models,
                    then you can set this argument to 5. In such a case, tpot will search
                    pipeline using only the top 5 algorithms/models that have been found
                    using .fit method.
                - if dictionary, then the keys should be the names of algorithms/models
                    and values shoudl be the parameters for each model/algorithm to be
                    optimized.
                - You can also set it to ``all`` consider all models available in
                    ai4water's Experiment module.
                - default is None, which means, the `tpot_config` argument will be None

            selection_criteria :
                The name of performance metric. If ``models`` is integer, then
                according to this performance metric the models will be choosen.
                By default the models will be selected based upon their mse values
                on test data.
            scoring : the performance metric to use for finding the pipeline.
            tpot_args :
                any keyword argument for tpot's Regressor_ or Classifier_ class.
                This can include arguments like ``generations``, ``population_size`` etc.

        Returns
        -------
            the tpot object

        Example
        -------
            >>> from ai4water.experiments import MLRegressionExperiments
            >>> from ai4water.datasets import busan_beach
            >>> exp = MLRegressionExperiments(exp_name=f"tpot_reg_{dateandtime_now()}")
            >>> exp.fit(data=busan_beach())
            >>> tpot_regr = exp.fit_with_tpot(busan_beach(), 2, generations=1, population_size=2)

        .. _tpot:
            http://epistasislab.github.io/tpot/

        .. _Regressor:
            http://epistasislab.github.io/tpot/api/#regression

        .. _Classifier:
            http://epistasislab.github.io/tpot/api/#classification
        """
        tpot_caller = self.tpot_estimator
        assert tpot_caller is not None, f"tpot must be installed"

        param_space = {}
        tpot_config = None
        for m in self.models:
            getattr(self, m)()
            ps = getattr(self, 'param_space')
            path = getattr(self, 'path')
            param_space[m] = {path: {p.name: p.grid for p in ps}}

        if isinstance(models, int):

            assert len(self.metrics)>1, f"""
            you must first run .fit() method in order to choose top {models} models"""

            # sort the models w.r.t their performance
            sorted_models = self.sort_models_by_metric(selection_criteria)

            # get names of models
            models = sorted_models.index.tolist()[0:models]

            tpot_config = {}
            for m in models:
                c: dict = param_space[f"{m}"]
                tpot_config.update(c)

        elif isinstance(models, list):

            tpot_config = {}
            for m in models:
                c: dict = param_space[f"model_{m}"]
                tpot_config.update(c)

        elif isinstance(models, dict):

            tpot_config = {}
            for mod_name, mod_paras in models.items():

                if "." in mod_name:
                    mod_path = mod_name
                else:
                    c: dict = param_space[f"model_{mod_name}"]
                    mod_path = list(c.keys())[0]
                d = {mod_path: mod_paras}
                tpot_config.update(d)

        elif isinstance(models, str) and models == "all":
            tpot_config = {}
            for mod_name, mod_config in param_space.items():
                mod_path = list(mod_config.keys())[0]
                mod_paras = list(mod_config.values())[0]
                tpot_config.update({mod_path: mod_paras})

        fname = os.path.join(self.exp_path, "tpot_config.json")
        with open(fname, 'w') as fp:
            json.dump(jsonize(tpot_config), fp, indent=True)

        tpot = tpot_caller(
            verbosity=self.verbosity+1,
            scoring=scoring,
            config_dict=tpot_config,
            **tpot_args
        )

        not_allowed_args = ["cross_validator", "wandb_config", "val_metric",
                            "loss", "optimizer", "lr", "epochs", "quantiles", "patience"]
        model_kws = self.model_kws
        for arg in not_allowed_args:
            if arg in model_kws:
                model_kws.pop(arg)

        dh = DataSet(data, **model_kws)
        train_x, train_y = dh.training_data()
        tpot.fit(train_x, train_y.reshape(-1, 1))

        visualizer = ProcessResults(path=self.exp_path)

        for idx, data_name in enumerate(['training', 'test']):

            x_data, y_data = getattr(dh, f"{data_name}_data")(key=str(idx))

            pred = tpot.fitted_pipeline_.predict(x_data)
            r2 = RegressionMetrics(y_data, pred).r2()

            # todo, perform inverse transform and deindexification
            visualizer.plot_results(
                pd.DataFrame(y_data.reshape(-1,)),
                pd.DataFrame(pred.reshape(-1,)),
                annotation_key='$R^2$', annotation_val=r2,
                show=self.verbosity,
                where='',  name=data_name
            )
        # save the python code of fitted pipeline
        tpot.export(os.path.join(self.exp_path, "tpot_fitted_pipeline.py"))

        # save each iteration
        fname = os.path.join(self.exp_path, "evaludated_individuals.json")
        with open(fname, 'w') as fp:
            json.dump(tpot.evaluated_individuals_, fp, indent=True)
        return tpot

    def _build_and_run(self,
        data,
        predict=True,
        view=False,
        title=None,
        cross_validate=False,
        **kwargs):

        """
        Builds and run one 'model' of the experiment.

        Since an experiment consists of many models, this method
        is also run many times.
        """
        model = self._build(title=title, **kwargs)

        self._fit(data=data, cross_validate=cross_validate)

        if view:
            model.view()

        if predict:
            return self._predict(data=data)

        # return the validation score
        return self._evaluate()

    def _build(self, title=None, **suggested_paras):
        """Builds the ai4water Model class"""

        suggested_paras = jsonize(suggested_paras)

        verbosity = max(self.verbosity-1, 0)
        if 'verbosity' in self.model_kws:
            verbosity = self.model_kws.pop('verbosity')

        model = Model(
            prefix=title,
            verbosity=verbosity,
            **self.model_kws,
            **suggested_paras
        )

        setattr(self, 'model_', model)
        return

    def _fit(self, data, cross_validate=False):
        """Trains the model"""

        if cross_validate:
            return self.model_.cross_val_score(data=data,
                scoring = self.model_.config['val_metric'])

        return self.model_.fit(data=data)

    def _evaluate(self, data="validation")->float:
        """Evaluates the model"""

        t, p = self.model_.predict(
            data=data,
            return_true=True,
            process_results=False)

        metrics = Metrics[self.mode](t, p,
            remove_zero=True, remove_neg=True,
            multiclass=self.model_.is_multiclass)

        test_metrics = {}
        for metric in self.monitor:
            test_metrics[metric] = getattr(metrics, metric)()
        self.model_iter_metric[self.iter_] = test_metrics
        self.iter_ += 1

        val_score = getattr(metrics, self.model_.val_metric)()

        if self.model_.config['val_metric'] in [
            'r2', 'nse', 'kge', 'r2_mod', 'r2_adj', 'r2_score', 'accuracy', 'f1_score']:
            val_score = 1.0 - val_score
        
        if not math.isfinite(val_score):
            val_score = 9999  # TODO, find a better way to handle this

        print(f"val_score: {val_score}")
        return val_score

    def _predict(self, data):
        """Makes predictions on training and test data from the model.
        It is supposed that the model has been trained before."""

        test_true, test_pred = self.model_.predict(data=data, return_true=True)

        train_true, train_pred = self.model_.predict(data='training', return_true=True)

        return (train_true, train_pred), (test_true, test_pred)


class TransformationExperiments(Experiments):
    """Helper class to conduct experiments with different transformations

        Example:
            >>> from ai4water.datasets import busan_beach
            >>> from ai4water.experiments import TransformationExperiments
            >>> from ai4water.hyperopt import Integer, Categorical, Real
            ... # Define your experiment
            >>> class MyTransformationExperiments(TransformationExperiments):
            ...
            ...    def update_paras(self, **kwargs):
            ...        _layers = {
            ...            "LSTM": {"config": {"units": int(kwargs['lstm_units']}},
            ...            "Dense": {"config": {"units": 1, "activation": kwargs['dense_actfn']}},
            ...            "reshape": {"config": {"target_shape": (1, 1)}}
            ...        }
            ...        return {'model': {'layers': _layers},
            ...                'lookback': int(kwargs['lookback']),
            ...                'batch_size': int(kwargs['batch_size']),
            ...                'lr': float(kwargs['lr']),
            ...                'transformation': kwargs['transformation']}
            >>> data = busan_beach()
            >>> inputs = ['tide_cm', 'wat_temp_c', 'sal_psu', 'air_temp_c', 'pcp_mm', 'pcp3_mm']
            >>> outputs = ['tetx_coppml']
            >>> cases = {'model_minmax': {'transformation': 'minmax'},
            ...         'model_zscore': {'transformation': 'zscore'}}
            >>> search_space = [
            ...            Integer(low=16, high=64, name='lstm_units', num_samples=2),
            ...            Integer(low=3, high=15, name="lookback", num_samples=2),
            ...            Categorical(categories=[4, 8, 12, 16, 24, 32], name='batch_size'),
            ...            Real(low=1e-6, high=1.0e-3, name='lr', prior='log', num_samples=2),
            ...            Categorical(categories=['relu', 'elu'], name='dense_actfn'),
            ...        ]
            >>> x0 = [20, 14, 12, 0.00029613, 'relu']
            >>> experiment = MyTransformationExperiments(cases=cases, input_features=inputs,
            ...                output_features=outputs, exp_name="testing"
            ...                 param_space=search_space, x0=x0)
    """

    @property
    def mode(self):
        return "regression"

    def __init__(self,
                 param_space=None,
                 x0=None,
                 cases: dict = None,
                 exp_name: str = None,
                 num_samples: int = 5, 
                 verbosity: int = 1,
                 **model_kws):
        self.param_space = param_space
        self.x0 = x0
        self.model_kws = model_kws

        exp_name = exp_name or 'TransformationExperiments' + f'_{dateandtime_now()}'

        super().__init__(cases=cases,
                         exp_name=exp_name,
                         num_samples=num_samples,
                         verbosity=verbosity)

    @property
    def tpot_estimator(self):
        return None

    def update_paras(self, **suggested_paras):
        raise NotImplementedError(f"""
You must write the method `update_paras` which should build the Model with suggested parameters
and return the keyword arguments including `model`. These keyword arguments will then
be used to build ai4water's Model class.
""")

    def _build(self, title=None, **suggested_paras):
        """Builds the ai4water Model class"""

        suggested_paras = jsonize(suggested_paras)

        verbosity = max(self.verbosity-1, 0)
        if 'verbosity' in self.model_kws:
            verbosity = self.model_kws.pop('verbosity')

        model = Model(
            prefix=title,
            verbosity=verbosity,
            **self.update_paras(**suggested_paras),
            **self.model_kws
        )

        setattr(self, 'model_', model)
        return

    def build_from_config(self, data, config_path, weight_file, **kwargs):

        model = Model.from_config_file(config_path=config_path)
        weight_file = os.path.join(model.w_path, weight_file)
        model.update_weights(weight_file=weight_file)

        model = self.process_model_before_fit(model)

        train_true, train_pred = model.predict(data=data, return_true=True)

        test_true, test_pred = model.predict(data='test', return_true=True)

        model.dh_.config['allow_nan_labels'] = 1
        model.predict()

        return (train_true, train_pred), (test_true, test_pred)

    def process_model_before_fit(self, model):
        """So that the user can perform procesisng of the model by overwriting this method"""
        return model


def sort_array(array):
    """
    array: [4, 7, 3, 9, 4, 8, 2, 8, 7, 1]
    returns: [4, 4, 3, 3, 3, 3, 2, 2, 2, 1]
    """
    results = np.array(array, dtype=np.float32)
    iters = range(1, len(results) + 1)
    return [np.min(results[:i]) for i in iters]


def consider_exclude(exclude: Union[str, list], 
        models, 
        models_to_filter: Union[list, dict] = None):

    if isinstance(exclude, str):
        exclude = [exclude]

    if exclude is not None:
        exclude = ['model_' + _model if not _model.startswith('model_') else _model for _model in exclude]
        for elem in exclude:
            assert elem in models, f"""
                {elem} to `exclude` is not available.
                Available models are {models} and you wanted to exclude
                {exclude}"""

            if models_to_filter is not None:
                assert elem in models_to_filter, f'{elem} is not in models'
                if isinstance(models_to_filter, list):
                    models_to_filter.remove(elem)
                else:
                    models_to_filter.pop(elem)
    return


def load_json_file(fpath):

    with open(fpath, 'r') as fp:
        result = json.load(fp)
    return result


def save_json_file(fpath, obj):

    with open(fpath, 'w') as fp:
        json.dump(jsonize(obj), fp)


def shred_model_name(model_name):
    key = model_name[6:] if model_name.startswith('model_') else model_name
    key = key[0:-9] if key.endswith("Regressor") else key
    return key


import os
from typing import Union

import numpy as np
import pandas as pd
from easy_mpl import imshow
import matplotlib.pyplot as plt

# TODO
# rank histogram, reliability diagram, ROC curve


class Plots(object):
    # TODO initialte this class with at least path

    def __init__(self, config, path=None, model=None):

        self.path = path or os.path.join(os.getcwd(), "results")
        self.ml_model = model
        self.config = config

    @property
    def training_data(self, *args, **kwargs):
        raise AttributeError

    @property
    def validation_data(self, *args, **kwargs):
        raise AttributeError

    @property
    def test_data(self, *args, **kwargs):
        raise AttributeError

    @property
    def in_cols(self):
        return self.config['input_features']

    @property
    def out_cols(self):
        return self.config['output_features']

    def _imshow_3d(self,
                   activation,
                   lyr_name,
                   xticklabels=None,
                   save=True,
                   where='activations',
                   xlabel=None):

        act_2d = []
        for i in range(activation.shape[2]):
            act_2d.append(activation[:, :, i])
        activation_2d = np.concatenate(act_2d, axis=1)
        self._imshow(activation_2d,
                     lyr_name + " Activations (3d of {})".format(activation.shape),
                     save,
                     lyr_name,
                     where=where,
                     xticklabels=xticklabels,
                     xlabel=xlabel)
        return

    def _imshow(self,
                img,
                label: str = '',
                save=True,
                fname=None,
                interpolation: str = 'none',
                where='',
                rnn_args=None,
                cmap=None,
                show=False,
                **kwargs):

        assert np.ndim(img) == 2, "can not plot {} with shape {} and ndim {}".format(label, img.shape, np.ndim(img))

        axis, im = imshow(img,
                          aspect="auto",
                          interpolation=interpolation,
                          cmap=cmap,
                          xlabel=kwargs.get('xlabel', 'inputs'),
                          show=False,
                          title=label)

        if rnn_args is not None:
            assert isinstance(rnn_args, dict)

            rnn_dim = int(img.shape[1] / rnn_args['n_gates'])
            [plt.axvline(rnn_dim * gate_idx - .5, linewidth=0.8, color='k')
             for gate_idx in range(1, rnn_args['n_gates'])]

            kwargs['xlabel'] = rnn_args['gate_names_str']
            if "RECURRENT" in label.upper():
                plt.ylabel("Hidden Units")
            else:
                plt.ylabel("Channel Units")
        else:
            axis.set_ylabel('Examples' if 'weight' not in label.lower() else '')
            xlabels = kwargs.get('xticklabels', None)
            if xlabels is not None:
                if len(xlabels) < 30:
                    axis.set_xticklabels(xlabels, rotation=90)

        plt.colorbar(im)
        self.save_or_show(save, fname, where=where, show=show)

        return

    def plot1d(self, array, label: str = '', show=False, fname=None, rnn_args=None, where=''):
        plt.close('all')
        plt.plot(array)
        plt.xlabel("Examples")
        plt.title(label)

        if rnn_args is not None:
            assert isinstance(rnn_args, dict)

            rnn_dim = int(array.shape[0] / rnn_args['n_gates'])
            [plt.axvline(rnn_dim * gate_idx - .5, linewidth=0.5, color='k')
             for gate_idx in range(1, rnn_args['n_gates'])]
            plt.xlabel(rnn_args['gate_names_str'])

        self.save_or_show(save=True, fname=fname, where=where, show=show)

        return

    def save_or_show(self, *args, **kwargs):

        return save_or_show(self.path, *args, **kwargs)

    def plot2d_act_for_a_sample(self, activations, sample=0, save: bool = False, name: str = None):
        from ai4water.utils.visualizations import init_subplots

        fig, axis = init_subplots(height=8)
        # for idx, ax in enumerate(axis):
        im = axis.imshow(activations[sample, :, :].transpose(), aspect='auto')
        axis.set_xlabel('lookback')
        axis.set_ylabel('inputs')
        axis.set_title('Activations of all inputs at different lookbacks for sample ' + str(sample))
        fig.colorbar(im)
        self.save_or_show(save=save, fname=name + '_' + str(sample), where='path')

        return

    def plot1d_act_for_a_sample(self, activations, sample=0, save=False, name=None):
        _, axis = plt.subplots()

        for idx in range(self.lookback-1):
            axis.plot(activations[sample, idx, :].transpose(), label='lookback '+str(idx))
        axis.set_xlabel('inputs')
        axis.set_ylabel('activation weight')
        axis.set_title('Activations at different lookbacks for all inputs for sample ' + str(sample))
        self.save_or_show(save=save, fname=name + '_' + str(sample), where='path')

        return

    def plot_train_data(self, how='3d', save=True,  **kwargs):

        x,  _ = self.training_data(**kwargs)
        self.plot_model_input_data(x, how=how, save=save, which='training')

        return

    def plot_val_data(self, how='3d', save=True,  **kwargs):

        x, y = self.validation_data(**kwargs)
        self.plot_model_input_data(x, how=how, save=save, which='validation')

        return

    def plot_test_data(self, how='3d', save=True,  **kwargs):

        x, _ = self.test_data(**kwargs)
        self.plot_model_input_data(x, how=how, save=save, which='test')

        return

    def plot_model_input_data(self,
                              in_data: Union[list, np.ndarray],
                              how: str,
                              save: bool,
                              which: str = 'training'
                              ) -> None:

        assert how in ['3d', 'hist']

        if not isinstance(in_data, list):
            if isinstance(in_data, dict):
                in_data = list(in_data.values())
            else:
                assert isinstance(in_data, np.ndarray)
                in_data = [in_data]

        for idx, inputs in enumerate(in_data):
            if np.ndim(inputs) == 3:
                if how.upper() == "3D":
                    self._imshow_3d(inputs, which + '_data_' + str(idx), save=save, where='data')
            elif np.ndim(inputs) == 2:
                if how.upper() == "3D":
                    self._imshow(inputs, save=save, fname=which + '_data_' + str(idx), where='data')
                else:
                    self.plot_histogram(inputs,
                                        save=save,
                                        fname=which+'_data_' + str(idx),
                                        features=self.in_cols,
                                        where='data')
            else:
                print(f'skipping shape is {inputs.shape}')
        return

    def plot_histogram(self,
                       data: np.ndarray,
                       save: bool = True,
                       fname='hist',
                       features=None,
                       where='data'
                       ):

        assert data.ndim == 2

        data = pd.DataFrame(data, columns=features)
        data.hist(figsize=(12, 12))

        self.save_or_show(save=save, fname=fname+'hist', where=where)
        return


def validate_freq(df, freq):
    assert isinstance(df.index, pd.DatetimeIndex), "index of dataframe must be pandas DatetimeIndex"
    assert freq in ["weekly", "monthly",
                    "yearly"], f"freq must be one of {'weekly', 'monthly', 'yearly'} but it is {freq}"
    return


def _get_nrows_and_ncols(n_subplots, n_rows=None):
    if n_rows is None:
        n_rows = int(np.sqrt(n_subplots))
    n_cols = max(int(n_subplots / n_rows), 1)  # ensure n_cols != 0
    n_rows = int(n_subplots / n_cols)

    while not ((n_subplots / n_cols).is_integer() and
               (n_subplots / n_rows).is_integer()):
        n_cols -= 1
        n_rows = int(n_subplots / n_cols)
    return n_rows, n_cols


def save_or_show(path, save: bool = True, fname=None, where='', dpi=300, bbox_inches='tight', close=True,
                 show=False):

    if save:
        assert isinstance(fname, str)
        if "/" in fname:
            fname = fname.replace("/", "__")
        if ":" in fname:
            fname = fname.replace(":", "__")

        save_dir = os.path.join(path, where)

        if not os.path.exists(save_dir):
            assert os.path.dirname(where) in ['',
                                              'activations',
                                              'weights',
                                              'plots', 'data', 'results'], f"unknown directory: {where}"
            save_dir = os.path.join(path, where)

            if not os.path.exists(save_dir):
                os.makedirs(save_dir)

        fname = os.path.join(save_dir, fname + ".png")

        plt.savefig(fname, dpi=dpi, bbox_inches=bbox_inches)
    if show:
        plt.show()

    if close:
        plt.close('all')
    return


def to_1d_array(array_like) -> np.ndarray:

    if array_like.__class__.__name__ in ['list', 'tuple', 'Series']:
        return np.array(array_like)

    elif array_like.__class__.__name__ == 'ndarray':
        if array_like.ndim == 1:
            return array_like
        else:
            assert array_like.size == len(array_like), f'cannot convert multidim ' \
                                                       f'array of shape {array_like.shape} to 1d'
            return array_like.reshape(-1, )

    elif array_like.__class__.__name__ == 'DataFrame' and array_like.ndim == 2:
        return array_like.values.reshape(-1,)
    else:
        raise ValueError(f'cannot convert object array {array_like.__class__.__name__}  to 1d ')



import os
from typing import Union

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

try:
    import lime
except ModuleNotFoundError:
    lime = None

from ._explain import ExplainerMixin


class LimeExplainer(ExplainerMixin):

    """
    Wrapper around LIME module.

    Example:
        >>> from ai4water import Model
        >>> from ai4water.datasets import busan_beach
        >>> model = Model(model="GradientBoostingRegressor")
        >>> model.fit(data=busan_beach())
        >>> lime_exp = LimeExplainer(model=model,
        ...                       train_data=model.training_data()[0],
        ...                       data=model.test_data()[0],
        ...                       mode="regression")
        >>> lime_exp.explain_example(0)

    Attributes:
        explaination_objects : location explaination objects for each individual example/instance
    """
    def __init__(
            self,
            model,
            data,
            train_data,
            mode: str,
            explainer=None,
            path=None,
            feature_names: list = None,
            verbosity: Union[int, bool] = True,
            **kwargs
    ):
        """
        Arguments:
            model : the model to explain. The model must have `predict` method.
            data : the data to explain. This would typically be test data but it
                can be any data.
            train_data : the data on which the model was trained.
            mode : either of `regression` or `classification`
            explainer : The explainer to use. By default, LimeTabularExplainer is used.
            path : path where to save all the plots. By default, plots will be saved in
                current working directory.
            feature_names : name/names of features.
            verbosity : whether to print information or not.
        """
        self.model = model
        self.train_data = to_np(train_data)

        super(LimeExplainer, self).__init__(path=path or os.getcwd(), data=to_np(data), features=feature_names)

        self.mode = mode
        self.verbosity = verbosity
        self.explainer = self._get_explainer(explainer, **kwargs)

        self.explaination_objects = {}

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, x):
        if x is not None:
            assert x in ["regression", "classification"], f"mode must be either regression or classification not {x}"
        self._mode = x

    def _get_explainer(self, proposed_explainer=None, **kwargs):

        import lime.lime_tabular

        if proposed_explainer is None and self.data.ndim <= 2:
            lime_explainer = lime.lime_tabular.LimeTabularExplainer(
                self.train_data,
                feature_names=self.features,
                # class_names=['price'],
                # categorical_features=categorical_features,
                verbose=self.verbosity,
                mode=self.mode,
                **kwargs
            )
        elif proposed_explainer in lime.lime_tabular.__dict__.keys():
            lime_explainer = getattr(lime.lime_tabular, proposed_explainer)(
                self.train_data,
                feature_names=self.features,
                mode=self.mode,
                verbose=self.verbosity,
                **kwargs
            )
        elif self.data.ndim == 3:
            lime_explainer = lime.lime_tabular.RecurrentTabularExplainer(
                self.train_data,
                mode=self.mode,
                feature_names=self.features,
                verbose=self.verbosity,
                **kwargs
            )
        elif proposed_explainer is not None:
            lime_explainer = getattr(lime, proposed_explainer)(
                self.train_data,
                features=self.features,
                mode=self.mode,
                **kwargs
            )
        else:
            raise ValueError(f"Can not infer explainer. Please specify explainer to use.")
        return lime_explainer

    def __call__(self, *args, **kwargs):

        self.explain_all_examples(*args, **kwargs)

        return

    def explain_all_examples(self,
                             plot_type="pyplot",
                             name="lime_explaination",
                             num_features=None,
                             **kwargs
                             ):
        """
        Draws and saves plot for all examples of test_data.

        Arguments:
            plot_type :
            name :
            num_features :
            kwargs : any keyword argument for `explain_instance`

        An example here means an instance/sample/data point.
        """
        for i in range(len(self.data)):
            self.explain_example(i, plot_type=plot_type, name=f"{name}_{i}",
                                 num_features=num_features, **kwargs)
        return

    def explain_example(
            self,
            index: int,
            plot_type: str = "pyplot",
            name: str = "lime_explaination",
            num_features: int = None,
            colors=None,
            annotate=False,
            show=False,
            save=True,
            **kwargs
    )->plt.Figure:
        """
        Draws and saves plot for a single example of test_data.

        Arguments:
            index : index of test_data
            plot_type : either pyplot or html
            name : name with which to save the file
            num_features :
            colors :
            annotate : whether to annotate figure or not
            show : whether to show figure or not
            save : wheter to save figure or not
            kwargs : any keyword argument for `explain_instance`

        Returns:
            matplotlib figure if plot_type="pyplot" and show is False.
        """
        assert plot_type in ("pyplot", "html")

        exp = self.explainer.explain_instance(self.data[index],
                                              self.model.predict,
                                              num_features=num_features or len(self.features),
                                              **kwargs
                                              )

        self.explaination_objects[index] = exp

        fig = None
        if plot_type == "pyplot":
            plt.close()
            fig = as_pyplot_figure(exp, colors=colors, example_index=index, annotate=annotate)
            if save:
                plt.savefig(os.path.join(self.path, f"{name}_{index}"), bbox_inches="tight")
            if show:
                plt.show()
        else:
            exp.save_to_file(os.path.join(self.path, f"{name}_{index}"))

        return fig


def to_np(x) -> np.ndarray:

    if isinstance(x, pd.DataFrame):
        x = x.values
    else:
        assert isinstance(x, np.ndarray)

    return x


def as_pyplot_figure(
        inst_explainer,
        label=1,
        example_index=None,
        colors: [str, tuple, list] = None,
        annotate=False,
        **kwargs):
    """Returns the explanation as a pyplot figure.

    Will throw an error if you don't have matplotlib installed
    Args:
        inst_explainer : instance explainer
        label: desired label. If you ask for a label for which an
               explanation wasn't computed, will throw an exception.
               Will be ignored for regression explanations.
        colors : if tuple it must be names of two colors for +ve and -ve
        example_index :
        annotate : whether to annotate the figure or not?
        kwargs: keyword arguments, passed to domain_mapper

    Returns:
        pyplot figure (barchart).
    """
    textstr = f"""Prediction: {round(inst_explainer.predicted_value, 2)}
Local prediction: {round(inst_explainer.local_pred.item(), 2)}"""

    if colors is None:
        colors = ([0.9375, 0.01171875, 0.33203125], [0.23828125, 0.53515625, 0.92578125])
    elif isinstance(colors, str):
        colors = (colors, colors)

    exp = inst_explainer.as_list(label=label, **kwargs)
    fig = plt.figure()
    vals = [x[1] for x in exp]
    names = [x[0] for x in exp]
    vals.reverse()
    names.reverse()

    if isinstance(colors, tuple):
        colors = [colors[0] if x > 0 else colors[1] for x in vals]

    pos = np.arange(len(exp)) + .5
    h = plt.barh(pos, vals, align='center', color=colors)
    plt.yticks(pos, names)
    if inst_explainer.mode == "classification":
        title = 'Local explanation for class %s' % inst_explainer.class_names[label]
    else:
        title = f'Local explanation for example {example_index}'
    plt.title(title)
    plt.grid(linestyle='--', alpha=0.5)

    if annotate:
        # https://stackoverflow.com/a/59109053/5982232
        plt.legend(h, [textstr], loc="best",
                   fancybox=True, framealpha=0.7,
                   handlelength=0, handletextpad=0)
    return fig




import os
import json
import inspect
import warnings
from typing import Union
from copy import copy, deepcopy


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

from sklearn.model_selection import train_test_split
from sklearn.model_selection import KFold, LeaveOneOut, TimeSeriesSplit

import ai4water.datasets as datasets
from ai4water.datasets import all_datasets
from ai4water.utils.utils import TrainTestSplit
from ai4water.utils.plotting_tools import Plots
from ai4water.preprocessing.imputation import Imputation
from ai4water.utils.utils import prepare_data, jsonize, to_datetime_index, print_something

from .utils import check_for_classification
from .utils import consider_intervals, decode
from .utils import load_data_from_hdf5


cmap_cv = plt.cm.coolwarm

class _DataSet(Plots):

    def __init__(self, config, path=os.getcwd()):

        Plots.__init__(self, config=config, path=path)

    def training_data(self):
        raise NotImplementedError
    
    def validation_data(self):
        raise NotImplementedError
    
    def test_data(self):
        raise NotImplementedError
    
    def KFold_splits(self, n_splits=5):
        raise NotImplementedError
    
    def LeaveOneOut_Splits(self):
        raise NotImplementedError
    
    def TimeSeriesSplit_splits(self, n_splits=5):
        raise NotImplementedError
    
    @classmethod
    def from_h5(cls, h5_file: str):
        raise NotImplementedError
    
    def to_disk(self, path: str):
        raise NotImplementedError

    def return_xy(self, x, y, initial):

        if self.mode == "classification" and  self.is_binary:
            if len(y) == y.size:
                y = y.reshape(-1,1)

        if self.verbosity > 0:
            print(f"{'*' * 5} {initial} {'*' * 5}")
            print_something(x, "input_x")
            print_something(y, "target")

        return x, y

    def return_x_yy(self, x, prev_y, y, initial):

        if self.verbosity > 0:
            print(f"{'*' * 5} {initial} data {'*' * 5}")
            print_something(x, "input_x")
            print_something(prev_y, "prev_y")
            print_something(y, "target")
        return x, prev_y, y


class DataSet(_DataSet):
    """The purpose of DataSet is to convert unprepared/raw data into prepared data.
    A prepared data consists of x,y pairs where x is inputs and y is outputs. There
    are >1 examples in a DataSet. Both inputs and outputs consists of same number
    of examples. An example consists of one input, output pair which can be given
    to a supervised machine learning algorithm for training. For tabular data, the
    number of examples does not necessarily match number of rows. The number of
    examples depend upon multiple factors such as presence of intervals, how
    nans are handled and the arguments related to time series data preparation
    which are listed in detail in prepare_data function.

    DataSet class can accept the raw, unprepared data in a variety of formats such
    as .csv, .xlsx, .parquet, .mat, .n5 etc. For details see this. The DataSet
    class can save the prepared data into an hdf5 file which can susequently be
    used to load the data and save the time.

    Methods
    ------------
    - training_data: returns training data
    - validation_data: returns validation data
    - test_data: returns test data
    - from_h5:
    - to_disk
    - KFold_splits: creates splits using `KFold` of sklearn
    - LeaveOneOut_splits: creates splits using `LeaveOneOut` of sklearn
    - TimeSeriesSplit_splits: creates splits using `TimeSeriesSplit` of sklearn
    - total_exs

    """
    def __init__(
        self,
        data,
        input_features: Union[str, list]=None,
        output_features:Union[str, list] = None,
        dataset_args: dict = None,

        ts_args: dict = None,

        split_random: bool = False,
        train_fraction: float = 0.7,
        val_fraction: float = 0.2,
        indices: dict = None,

        intervals=None,
        shuffle: bool = True,
        allow_nan_labels: int = 0,
        nan_filler: dict = None,
        batch_size: int = 32,
        drop_remainder: bool = False,
        teacher_forcing: bool = False,
        allow_input_nans: bool =  False,

        seed: int = 313,
        verbosity: int = 1, 
        mode: str = None,
        category: str = None,
        save: bool = False
        ):
        """
        Initializes the DataSet class

        Parameters
        ----------
            data :
                source from which to make the data. It can be one of the following:

                - pandas dataframe: each columns is a feature and each row is an example
                - numpy array
                - xarray dataset: it can be xarray dataset
                - path like: if the path is the path of a file, then this file can
                    be a csv/xlsx/nc/npz/mat/parquet/feather file. The .nc file
                    will be read using xarray to load datasets. If the path refers
                    to a directory, it is supposed that each file in the directory refers to one example.
                - ai4water dataset : name of any of dataset name from ai4water.datasets
                - name of .h5 file
            input_features : Union[list, dict, str, None]
                features to use as input. If `data` is pandas dataframe
                then this is list of column names from `data` to be used as input.
            output_features : Union[list, dict, str, None]
                features to use as output. When `data` is dataframe
                then it is list of column names from `data` to be used as output.
                If `data` is `dict`, then it must be consistent with `data`.
                Default is None,which means the last column of data will be
                used as output. In case of multi-class classification, the output
                column is not supposed to be one-hot-encoded rather in the form
                of [0,1,2,0,1,2,1,2,0] for 3 classes. One-hot-encoding is done
                inside the model.
            dataset_args : dict
                additional arguments for AI4Water's [datasets][ai4water.datasets]
            ts_args : dict, optional
                This argument should only be used if the data is time series data.
                It must be a dictionary which is then passed to :py:func:`ai4water.utils.prepare_data`
                for data preparation. Possible keys in dictionay are:
                    - lookback
                    - forecast_len
                    - forecast_step
                    - input_steps
            split_random : bool, optional
                whether to split the data into training and test randomly or not.
            train_fraction : float
                Fraction of the complete data to be used for training
                purpose. Must be greater than 0.0.
            val_fraction : float
                The fraction of the training data to be used for validation.
                Set to 0.0 if no validation data is to be used.
            indices : dict, optional
                A dictionary with two possible keys, 'training', 'validation'.
                It determines the indices to be used to select training, validation
                and test data. If indices are given for training, then train_fraction
                must not be given. If indices are given for validation, then indices
                for training must also be given and  val_fraction must not be given.
                Therefore, the possible keys in indices dictionary are follwoing 
                    - 'training'
                    - 'training' and 'validation'
            intervals :
                tuple of tuples where each tuple consits of two integers, marking
                the start and end of interval. An interval here means indices
                from the data. Only rows within those indices will be used when preparing
                data/batches for NN. This is handy when our input data
                contains chunks of missing values or when we don't want to consider several
                rows in input data during data_preparation.
                For further usage see `examples/using_intervals`
            shuffle : bool
            allow_nan_labels : bool
                whether to allow examples with nan labels or not.
                if it is > 0, and if target values contain Nans, those examples
                will not be ignored and will be used as it is.
                In such a case a customized training and evaluation
                step is performed where the loss is not calculated for predictions
                corresponding to nan observations. Thus this option can be useful
                when we are predicting more than 1 target and some of the examples
                have some of their labels missing. In such a scenario, if we set this
                option to >0, we don't need to ignore those samples at all during data
                preparation. This option should be set to > 0 only when using tensorflow
                for deep learning models. if == 1, then if an example has label [nan, 1]
                it will not be removed while the example with label [nan, nan]
                will be ignored/removed. If ==2, both examples (mentioned before) will be
                considered/will not be removed. This means for multi-outputs, we can end
                up having examples whose all labels are nans. if the number of outputs
                are just one. Then this must be set to 2 in order to use samples with nan labels.
            nan_filler : dict
                This argument determines the imputation technique used to fill the nans in
                the data. The imputation is actually performed by :py:class:`ai4water.preprocessing.Imputation`
                class. Therefore this argument determines the interaction with `Imputation` class.
                The default value is None, which will raise error if missing/nan values
                are encountered in the input data. The user can however specify a
                dictionary whose one key must be `method`. The value of 'method'
                key can be `fillna` or `interpolate`.  For example, to do forward
                filling, the user can do as following

                >>> {'method': 'fillna', 'imputer_args': {'method': 'ffill'}}

                For details about fillna keyword options see fillna_

                For `interpolate`, the user can specify  the type of interpolation
                for example

                >>> {'method': 'interpolate', 'imputer_args': {'method': 'spline', 'order': 2}}

                will perform spline interpolation with 2nd order.
                For other possible options/keyword arguments for interpolate_
                [see]()
                The filling or interpolation is done columnwise, however, the user
                can specify how to do for each column by providing the above mentioned
                arguments as dictionary or list. The sklearn based imputation methods
                can also be used in a similar fashion. For KNN

                >>> {'method': 'KNNImputer', 'imputer_args': {'n_neighbors': 3}}

                or for iterative imputation

                >>> {'method': 'IterativeImputer', 'imputer_args': {'n_nearest_features': 2}}

                To pass additional arguments one can make use of `imputer_args`
                keyword argument

                >>> {'method': 'KNNImputer', 'features': ['b'], 'imputer_args': {'n_neighbors': 4}},

                For more on sklearn based imputation methods see this blog_
            batch_size : int
                size of one batch. Only relevent if `drop_remainder` is True.
            drop_remainder : bool
                whether to drop the remainder if len(data) % batch_size != 0 or not?
            teacher_forcing : bool
                whether to return previous output/target/ground
                truth or not. This is useful when the user wants to feed output
                at t-1 as input at timestep t. For details about this technique
                see this article_
            allow_input_nans : bool, optional
                If False, the examples containing nans in inputs will be removed.
                Setting this to True will result in feeding nan containing data
                to your algorithm unless nans are filled with `nan_filler`.
            seed : int
                random seed for reproducibility
            verbosity : int
            mode : str
            category : str
            save : bool
                whether to save the data in an h5 file or not.


        Example
        -------
            >>> import pandas as pd
            >>> import numpy as np
            >>> from ai4water.preprocessing import DataSet
            >>> data_ = pd.DataFrame(np.random.randint(0, 1000, (50, 2)), columns=['input', 'output'])
            >>> data_set = DataSet(data=data_, ts_args={'lookback':5})
            >>> x,y = data_set.training_data()

        .. _fillna:
            https://pandas.pydata.org/pandas-docs/version/0.22.0/generated/pandas.DataFrame.fillna.html
        .. _article:
            https://machinelearningmastery.com/teacher-forcing-for-recurrent-neural-networks/
        .. _interpolate:
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.interpolate.html
        .. _blog:
            https://scikit-learn.org/stable/auto_examples/impute/plot_missing_values.html#sphx-glr-auto-examples-impute-plot-missing-values-py

        Note
        ----
        The word 'index' is not allowed as column name, input_features or output_features
        """

        indices = indices or {}

        if indices:
            assert split_random is False, "indices cannot be used with split_random"
        
        if 'training' in indices:
            assert train_fraction == 0.7, f"""
            You can not set training data using both indices and train_fraction.
            Use either indices or train_fraction."""
    
        if 'validation' in indices:
            assert val_fraction == 0.2, f"""
                You can not set validation data using both indices and val_fraction. 
                Use either indices or val_fraction."""
            assert 'training' in indices, f"""
            when defining validation data using indices, training data must also be 
            defined using indices."""
        
        assert val_fraction < 1.0, f"""
            val_fraction must be less than 1.0 but it is {val_fraction}.
            """

        self.dataset_args = dataset_args

        self.config = {
            'input_features': input_features,
            'output_features': output_features
        }
        self.nan_filler = nan_filler

        self.data = self._process_data(
            data,
            input_features,
            output_features)

        self.ts_args = ts_args
        self.split_random = split_random
        self.indices = indices
        self.train_fraction = train_fraction
        self.val_fraction = val_fraction
        self.shuffle = shuffle
        self.batch_size = batch_size

        self.intervals = intervals
        self.allow_nan_labels=allow_nan_labels
        self.teacher_forcing = teacher_forcing
        self.drop_remainder = drop_remainder
        self.allow_input_nans = allow_input_nans

        self.verbosity = verbosity
        self.seed = seed
        self.mode = mode
        self.category = category
        self.save = save

        self.scalers = {}
        self.indexes = {}
        self.index_types = {}

        self._input_features = copy(input_features)

        if save:
            self.to_disk()

        _DataSet.__init__(self, config=self.config, path=os.getcwd())

    def init_paras(self) -> dict:
        """Returns the initializing parameters of this class"""
        signature = inspect.signature(self.__init__)

        init_paras = {}
        for para in signature.parameters.values():
            init_paras[para.name] = getattr(self, para.name)

        return init_paras

    @property
    def ts_args(self):
        return self._ts_args

    @ts_args.setter
    def ts_args(self, _ts_args:dict = None):
        default_args = {'input_steps': 1,
                        'lookback': 1,
                        'forecast_len':  1,
                        'forecast_step':  0,
                        'known_future_inputs': False
                        }
        
        if _ts_args:
            default_args.update(_ts_args)
        
        self._ts_args = default_args

    @property
    def lookback(self):
        return self.ts_args['lookback']

    @property
    def classes(self):
        _classes = []
        if self.mode == 'classification':
            if self.num_outs == 1:  # for binary/multiclass
                array = self.data[self._output_features].values
                _classes = np.unique(array[~np.isnan(array)])
            else:  # for one-hot encoded
                _classes = self._output_features

        return _classes

    @property
    def num_classes(self):
        return len(self.classes)

    @property
    def is_binary(self) -> bool:
        """Returns True if the porblem is binary classification"""
        _default = False
        if self.mode == 'classification':
            if self.num_outs == 1:
                array = self.data[self._output_features].values
                unique_vals = np.unique(array[~np.isnan(array)])
                if len(unique_vals) == 2:
                    _default = True
            else:
                pass  # todo, check when output columns are one-hot encoded

        return _default

    @property
    def is_multiclass(self) -> bool:
        """Returns True if the porblem is multiclass classification"""
        _default = False
        if self.mode == 'classification':
            if self.num_outs == 1:
                array = self.data[self._output_features].values
                unique_vals = np.unique(array[~np.isnan(array)])
                if len(unique_vals) > 2:
                    _default = True
            else:
                pass  # todo, check when output columns are one-hot encoded

        return _default

    @property
    def is_multilabel(self) -> bool:
        """Returns True if the porblem is multilabel classification"""
        _default = False
        if self.mode == 'classification':
            if self.num_outs > 1:
                _default = True

        return _default

    @property
    def _to_categorical(self):
        # whether we have to convert y into one-hot encoded form
        _defualt = False

        if self.is_binary or self.is_multiclass:
            if self.num_outs == 1:
                _defualt = True
        # it seems sklearn can accept one-hot-encoded targets but xgb, lgbm and catboost can't
        # but since since sklearn can also accept non-one-hot-encoded targets for multiclass
        # let's not one-hot-encode for all ML algos
        if self.category == 'ML':
            _defualt = False

        return _defualt

    @property
    def teacher_forcing(self):
        return self._teacher_forcing

    @teacher_forcing.setter
    def teacher_forcing(self, x):
        self._teacher_forcing = x

    @property
    def input_features(self):
        _inputs = self.config['input_features']

        if _inputs is None and self.data is not None:
            assert isinstance(self.data, pd.DataFrame)
            _inputs = self.data.columns[0:-1].to_list()

        return _inputs

    @property
    def output_features(self):
        """for external use"""
        _outputs = self.config['output_features']

        if _outputs is None and self.data is not None:
            #assert isinstance(self.data, pd.DataFrame)
            if self.data.ndim == 2:
                _outputs = [col for col in self.data.columns if col not in self.input_features]
            else:
                _outputs = []  # todo
        return _outputs

    @property
    def _output_features(self):
        """for internal use"""
        _outputs = deepcopy(self.config['output_features'])

        if isinstance(self.data, list):
            assert isinstance(_outputs, list)

        elif isinstance(self.data, dict):
            assert isinstance(_outputs, dict), f"""
            data is of type dict while output_features are
            of type {_outputs.__class__.__name__}"""
            for k in self.data.keys():
                if k not in _outputs:
                    _outputs[k] = []

        elif _outputs is None and self.data is not None:
            assert isinstance(self.data, pd.DataFrame)
            _outputs = [col for col in self.data.columns if col not in self.input_features]

        return _outputs

    @property
    def num_ins(self):
        return len(self.input_features)
    
    @property
    def num_outs(self):
        return len(self.output_features)

    @property
    def batch_dim(self):

        default = "3D"
        if self.ts_args['lookback'] == 1:
            default = "2D"

        return default

    def _process_data(self,
                      data,
                      input_features,
                      output_features
                      ):

        if isinstance(data, str):
            _source = self._get_data_from_str(data, input_features, output_features)
            if isinstance(_source, str) and _source.endswith('.h5'):
                self._from_h5 = True

        elif isinstance(data, pd.DataFrame):
            _source = self._get_data_from_df(data, input_features, output_features)

        elif isinstance(data, np.ndarray):
            _source = self._get_data_from_ndarray(data, input_features, output_features)

        elif data.__class__.__name__ == "Dataset":
            _source = data

        elif isinstance(data, list):
            raise ValueError(f"""
            data is given as a list. For such cases either use DataSetUnion
            or DataSetPipeline insteadd of DataSet class""")

        elif isinstance(data, dict):
            raise ValueError(f"""
            data is given as a dictionary. For such cases either use DataSetUnion
            or DataSetPipeline insteadd of DataSet class""")

        elif data is None:
            return data

        else:
            assert data is not None
            raise ValueError(f"""
            unregnizable source of data of type {data.__class__.__name__} given
            """)

        _source = self.impute(_source)

        return _source

    def _get_data_from_ndarray(self, data, input_features, output_features):
        if data.ndim == 2:
            # if output_features is not defined, consider 1 output and name it
            # as 'output'
            if output_features is None:
                output_features = ['outout']
                self.config['output_features'] = output_features  # we should put it in config as well
            elif isinstance(output_features, str):
                output_features = [output_features]
            else:
                assert isinstance(output_features, list)

            if input_features is None:  # define dummy names for input_features
                input_features = [f'input_{i}' for i in range(data.shape[1]-len(output_features))]
                self.config['input_features'] = input_features

            return pd.DataFrame(data, columns=input_features + output_features)
        else:
            return data

    def _get_data_from_df(self, data, input_features, output_features):

        if input_features is None and output_features is not None:
                if isinstance(output_features, str):
                    output_features = [output_features]
                assert isinstance(output_features, list)
                input_features = [col for col in data.columns if col not in output_features]
                # since we have inferred the input_features, they should be put
                # back into config
                self.config['input_features'] = input_features
        
        return data

    def _get_data_from_str(self, data, input_features, output_features):
        if isinstance(output_features, str):
            output_features = [output_features]

        # dir path/file path/ ai4water dataset name
        if data.endswith('.h5'):
            _source = data
        if data.endswith('.csv'):
            _source = pd.read_csv(data)
            if _source.columns[0] in ['index', 'time', 'date']:
                _source.index = pd.to_datetime(_source.pop('index'))

        elif data.endswith('.xlsx') or data.endswith('xlx'):
            _source = pd.read_excel(data)
            if _source.columns[0] in ['index', 'time', 'date']:
                _source.index = pd.to_datetime(_source.pop('index'))

        elif data.endswith('.parquet'):
            _source = pd.read_parquet(data)

        elif data.endswith('.feather'):
            _source = pd.read_feather(data)
            if _source.columns[0] in ['index', 'time', 'date']:
                _source.index = pd.to_datetime(_source.pop('index'))

        # netcdf file
        elif data.endswith('.nc'):
            import xarray as xr
            _source = xr.open_dataset(data)
            _source = _source.to_dataframe()

        elif data.endswith('npz'):
            data = np.load(data)
            assert len(data) == 1
            d = []
            for k, v in data.items():
                d.append(v)

            data: np.ndarray = d[0]
            _source = pd.DataFrame(data, columns=input_features + output_features)

        # matlab's mat file
        elif data.endswith('.mat'):
            import scipy
            mat = scipy.io.loadmat(data)
            data: np.ndarray = mat['data']
            _source = pd.DataFrame(data, columns=input_features + output_features)

        elif os.path.isfile(data):
            assert os.path.exists(data)
            _source = data
        elif os.path.isdir(data):
            assert len(os.listdir(data)) > 1
            # read from directory
            raise NotImplementedError
        elif data in all_datasets:
            _source = self._get_data_from_ai4w_datasets(data)
        else:
            raise ValueError(f"unregnizable source of data given {data}")

        return _source

    def _get_data_from_ai4w_datasets(self, data):

        Dataset = getattr(datasets, data)

        dataset = Dataset()
        dataset_args = self.dataset_args
        if dataset_args is None:
            dataset_args = {}

        # if self.config['input_features'] is not None:

        dynamic_features = self.input_features + self.output_features

        data = dataset.fetch(dynamic_features = dynamic_features,
                             **dataset_args)

        data = data.to_dataframe(['time', 'dynamic_features']).unstack()

        data.columns = [a[1] for a in data.columns.to_flat_index()]

        return data

    def impute(self, data):
        """Imputes the missing values in the data using `Imputation` module"""
        if self.nan_filler is not None:

            if isinstance(data, pd.DataFrame):

                _source = self._impute(data, self.nan_filler)

            else:
                raise NotImplementedError
        else:
            _source = data

        return _source

    def _impute(self, data, impute_config):

        if isinstance(impute_config, str):
            method, impute_args = impute_config, {}
            data = Imputation(data, method=method, **impute_args)()

        elif isinstance(impute_config, dict):
            data = Imputation(data, **impute_config)()

        elif isinstance(impute_config, list):
            for imp_conf in impute_config:
                data = Imputation(data, **imp_conf)()

        else:
            raise NotImplementedError(f'{impute_config.__class__.__name__}')

        return data

    def get_indices(self):
        """If the data is to be divded into train/test based upon indices,
        here we create train_indices and test_indices. The train_indices
        contain indices for both training and validation data.
        """

        tot_obs = self.total_exs(**self.ts_args)

        all_indices = np.arange(tot_obs)

        if len(self.indices) == 0 :
            if self.train_fraction < 1.0:
                if self.split_random:
                    train_indices, test_indices = train_test_split(
                        all_indices,
                        train_size=self.train_fraction,
                        random_state=self.seed
                        )
                else:
                    train_indices, test_indices = self._get_indices_by_seq_split(
                        all_indices,
                        self.train_fraction)
            else:  # no test data
                train_indices, test_indices = all_indices, []
        else:
            _train_indices = self.indices.get('training', None)
            _val_indices = self.indices.get('validation', None)
            _test_indices = self.indices.get('test', None)

            if _train_indices is not None:
                if _val_indices is None:
                    # even if val_fraction is > 0.0, we will separate validation 
                    # data from training later
                    _val_indices = np.array([])   # no validation set
                else:
                    assert isinstance(np.array(_val_indices), np.ndarray)
                    _val_indices = np.array(_val_indices)

                overlap = np.intersect1d(_train_indices, _val_indices)
                assert len(overlap) == 0, f"""
                    Training and validation indices must be mutually exclusive.
                    They contain {len(overlap)} overlaping values."""
                train_indices = np.sort(np.hstack([_train_indices, _val_indices]))

                if _test_indices is None:
                    # get test_indices by subtracting train_indices from all indices
                    test_indices = [ind for ind in all_indices if ind not in train_indices]
                    #_val_indices = np.array([])
                
            else:  # todo
                train_indices = []
                
        setattr(self, 'train_indices', train_indices)
        setattr(self, 'test_indices', test_indices)

        return np.array(train_indices).astype("int32"), np.array(test_indices).astype("int32")

    def _get_indices_by_seq_split(
        self, 
        all_indices:Union[list, np.ndarray], 
        train_fraction):
        """ sequential train/test split"""

        train_indices = all_indices[0:int(train_fraction * len(all_indices))]
        test_indices = all_indices[int(train_fraction * len(all_indices)):]
        return train_indices, test_indices


    def _training_data(self, key="_training", **kwargs):
        """training data including validation data"""

        train_indices, test_indices = self.get_indices()

        if 'validation' in self.indices:
            # when validation indices are given, we first prepare
            # complete data which contains training, validation and test data
            # TODO this is agains function definition
            indices = np.sort(np.hstack([train_indices, test_indices]))
        else:
            indices = train_indices

        data = self.data.copy()

        # numpy arrays are not indexed and is supposed that the whole array is
        # use as input
        if not isinstance(data, np.ndarray):
            data = self.indexify(data, key)

        # get x,_y, y
        x, prev_y, y = self._make_data(
            data,
            intervals=self.intervals,
            indices= indices,
            **kwargs)

        if not isinstance(self.data, np.ndarray):
            x, self.indexes[key] = self.deindexify(x, key)

        if self.mode == 'classification':
            y = check_for_classification(y, self._to_categorical)

        return x, prev_y, y

    def training_data(self, key="train", **kwargs):
        """training data excluding validation data"""

        if getattr(self, '_from_h5', False):
            return load_data_from_hdf5('training_data', self.data)

        x, prev_y, y = self._training_data(key=key, **kwargs)
        if self.val_fraction>0.0:
            # when no output is generated, corresponding index will not be saved
            idx = self.indexes.get(key, np.arange(len(x)))  # index also needs to be split
            x, prev_y, y, idx = self._train_val_split(x, prev_y, y, idx, 'training')

            # if drop remainder, we need to
            x, prev_y, y = self.check_for_batch_size(x, prev_y, y)

            self.indexes[key] = idx[0:len(x)]

        if self.teacher_forcing:
            return self.return_x_yy(x, prev_y, y, "Training")

        return self.return_xy(x, y, "Training")

    def validation_data(self, key="val", **kwargs):
        """validation data"""
        
        if getattr(self, '_from_h5', False):
            return load_data_from_hdf5('validation_data', self.data)

        x, prev_y, y = self._training_data(key=key, **kwargs)

        if self.val_fraction>0.0:
            idx = self.indexes.get(key, np.arange(len(x)))
            x, prev_y, y, idx = self._train_val_split(x, prev_y, y, idx, 'validation')

            x, prev_y, y = self.check_for_batch_size(x, prev_y, y)

            self.indexes[key] = idx[0:len(x)]
        else:
            x, prev_y, y = np.empty(0), np.empty(0), np.empty(0)

        if self.teacher_forcing:
            return self.return_x_yy(x, prev_y, y, "Validation")

        return self.return_xy(x, y, "Validation")

    def _train_val_split(self, x, prev_y, y, idx, return_type):
        """split x,y,idx,prev_y into training and validation data"""

        if self.split_random:
            # split x,y randomly
            splitter = TrainTestSplit(test_fraction=self.val_fraction, seed=self.seed)
            train_x, val_x, train_y, val_y = splitter.split_by_random(x, y)
            splitter = TrainTestSplit(test_fraction=self.val_fraction, seed=self.seed)
            train_idx, val_idx, train_prev_y, val_prev_y = splitter.split_by_random(
                idx, prev_y)

        elif 'validation' in self.indices:
            # separate indices were provided for validation data
            # it must be remembered that x,y now contains training+validation+test data
            # but based upon indices, we will choose either training or validation data
            val_indices = self.indices['validation']
            _train_indices, _ = self.get_indices()
            train_indices = [i for i in _train_indices if i not in val_indices]
            splitter = TrainTestSplit(train_indices=train_indices, test_indices=val_indices)
            train_x, val_x, train_y, val_y = splitter.split_by_indices(
                x, y
            )
            splitter = TrainTestSplit(train_indices=train_indices, test_indices=val_indices)
            train_idx, val_idx, train_prev_y, val_prev_y = splitter.split_by_indices(
                idx, prev_y)
        else:
            # split x,y sequentially
            splitter = TrainTestSplit(test_fraction=self.val_fraction)
            train_x, val_x, train_y, val_y = splitter.split_by_slicing(x, y)
            splitter = TrainTestSplit(test_fraction=self.val_fraction)
            train_idx, val_idx, train_prev_y, val_prev_y = splitter.split_by_slicing(idx, prev_y)
        
        if return_type=="training":
            return train_x, train_prev_y, train_y, train_idx
        
        return val_x, val_prev_y, val_y, val_idx

    def test_data(self, key="test", **kwargs):
        """test data"""
        if getattr(self, '_from_h5', False):
            return load_data_from_hdf5('test_data', self.data)

        if self.train_fraction < 1.0:

            data = self.data.copy()

            # numpy arrays are not indexed and is supposed that the whole array
            # is use as input
            if not isinstance(data, np.ndarray):
                data = self.indexify(data, key)

            _, test_indices = self.get_indices()

            if len(test_indices) > 0:  # it is possible that training and validation
                # indices cover whole data
                # get x,_y, y
                x, prev_y, y = self._make_data(
                    data,
                    intervals=self.intervals,
                    indices=test_indices,
                    **kwargs)

                x, prev_y, y = self.check_for_batch_size(x, prev_y, y)

                if not isinstance(self.data, np.ndarray):
                    x, self.indexes[key] = self.deindexify(x, key)

                if self.mode == 'classification':
                    y = check_for_classification(y, self._to_categorical)
            else:
                x, prev_y, y = np.empty(0), np.empty(0), np.empty(0)    
        else:
            x, prev_y, y = np.empty(0), np.empty(0), np.empty(0)

        if self.teacher_forcing:
            return self.return_x_yy(x, prev_y, y, "Test")

        return self.return_xy(x, y, "Test")

    def check_for_batch_size(self, x, prev_y=None, y=None):

        if self.drop_remainder:

            assert isinstance(x, np.ndarray)
            remainder = len(x) % self.batch_size

            if remainder:

                x = x[0:-remainder]

                if prev_y is not None:
                    prev_y = prev_y[0:-remainder]
                if y is not None:
                    y = y[0:-remainder]

        return x, prev_y, y

    def check_nans(self, data, input_x, input_y, label_y):
        """Checks whether anns are present or not and checks shapes of arrays
        being prepared."""
        if isinstance(data, pd.DataFrame):
            nans = data[self.output_features].isna()
            nans = nans.sum().sum()
            data = data.values
        else:
            nans = np.isnan(data[:, -self.num_outs:])
            # df[self.out_cols].isna().sum()
            nans = int(nans.sum())
        if nans > 0:
            if self.allow_nan_labels == 2:
                if self.verbosity > 0: print("""
                \n{} Allowing NANs in predictions {}\n""".format(10 * '*', 10 * '*'))
            elif self.allow_nan_labels == 1:
                if self.verbosity>0: print("""
                \n{} Ignoring examples whose all labels are NaNs {}\n
                """.format(10 * '*', 10 * '*'))
                idx = ~np.array([all([np.isnan(x) for x in label_y[i]]) for i in range(len(label_y))])
                input_x = input_x[idx]
                input_y = input_y[idx]
                label_y = label_y[idx]
                if int(np.isnan(data[:, -self.num_outs:][0:self.lookback]).sum() / self.num_outs) >= self.lookback:
                    self.nans_removed_4m_st = -9999
            else:
                if self.verbosity > 0:
                    print('\n{} Removing Examples with nan in labels  {}\n'.format(10 * '*', 10 * '*'))
                if self.num_outs == 1:
                    # find out how many nans were present from start of data until
                    # lookback, these nans will be removed
                    self.nans_removed_4m_st = np.isnan(data[:, -self.num_outs:][0:self.lookback]).sum()
                # find out such labels where 'y' has at least one nan
                nan_idx = np.array([np.any(i) for i in np.isnan(label_y)])
                non_nan_idx = np.invert(nan_idx)
                label_y = label_y[non_nan_idx]
                input_x = input_x[non_nan_idx]
                input_y = input_y[non_nan_idx]

                assert np.isnan(label_y).sum() < 1, """
                label still contains {} nans""".format(np.isnan(label_y).sum())

        assert input_x.shape[0] == input_y.shape[0] == label_y.shape[0], """
        shapes are not same"""

        if not self.allow_input_nans:
            assert np.isnan(input_x).sum() == 0, """input still contains {} nans
            """.format(np.isnan(input_x).sum())

        return input_x, input_y, label_y

    def indexify(self, data: pd.DataFrame, key):

        data = data.copy()
        dummy_index = False
        # for dataframes
        if isinstance(data.index, pd.DatetimeIndex):
            index = list(map(int, np.array(data.index.strftime('%Y%m%d%H%M'))))
            # datetime index
            self.index_types[key] = 'dt'
            original_index = pd.Series(index, index=index)
        else:
            try:
                index = list(map(int, np.array(data.index)))
                self.index_types[key] = 'int'
                original_index = pd.Series(index, index=index)
            except ValueError:  # index may not be convertible to integer, it may be
                # string values
                dummy_index = np.arange(len(data), dtype=np.int64)
                original_index = pd.Series(data.index, index=dummy_index)
                index = dummy_index
                self.index_types[key] = 'str'
                self.indexes[key] = {'dummy': dummy_index,
                                     'original': original_index}
        # pandas will add the 'datetime' column as first column.
        # This columns will only be used to keep
        # track of indices of train and test data.
        data.insert(0, 'index', index)

        self._input_features = ['index'] + self.input_features
        # setattr(self, 'input_features', ['index'] + self.input_features)
        self.indexes[key] = {'index': index, 'dummy_index': dummy_index,
                             'original': original_index}
        return data

    def deindexify(self, data:np.ndarray, key):

        _data, _index = self.deindexify_nparray(data, key)

        if self.indexes[key].get('dummy_index', None) is not None:
            _index = self.indexes[key]['original'].loc[_index].values

        if self.index_types[key] == 'dt':
            _index = to_datetime_index(_index)
        return _data, _index

    def get_batches(self, data):

        if self.batch_dim == "2D":
            return self.get_2d_batches(data)

        else:
            return self.check_nans(data, *prepare_data(data,
                                                       num_outputs=self.num_outs,
                                                       **self.ts_args))

    def get_2d_batches(self, data):
        # need to count num_ins based upon _input_features as it consider index
        num_ins = len(self._input_features)

        if not isinstance(data, np.ndarray):
            if isinstance(data, pd.DataFrame):
                data = data.values
            else:
                raise TypeError(f"unknown data type {data.__class__.__name__} for data ")

        if self.num_outs > 0:
            input_x = data[:, 0:num_ins]
            input_y, label_y = data[:, -self.num_outs:], data[:, -self.num_outs:]
        else:
            dummy_input_y = np.random.random((len(data), self.num_outs))
            dummy_y = np.random.random((len(data), self.num_outs))
            input_x, input_y, label_y = data[:, 0:num_ins], dummy_input_y, dummy_y

        assert self.lookback == 1, """
        lookback should be one for MLP/Dense layer based model, but it is {}
            """.format(self.lookback)

        return self.check_nans(data, input_x, input_y, np.expand_dims(label_y, axis=2))

    def _make_data(self, data, indices=None, intervals=None, shuffle=False):

        #if indices is not None:
        #    indices = np.array(indices).astype("int32")
            #assert isinstance(np.array(indices), np.ndarray), "indices must be array like"

        if isinstance(data, pd.DataFrame):
            data = data[self._input_features + self.output_features].copy()
            df = data
        else:
            data = data.copy()
            df = data

        if intervals is None:
            x, prev_y, y = self.get_batches(df)

            if indices is not None:
                # if indices are given then this should be done after `get_batches`
                # method
                x = x[indices]
                prev_y = prev_y[indices]
                y = y[indices]
        else:
            xs, prev_ys, ys = [], [], []
            for _st, _en in intervals:
                df1 = data[_st:_en]

                if df1.shape[0] > 0:

                    x, prev_y, y = self.get_batches(df1.values)

                    xs.append(x)
                    prev_ys.append(prev_y)
                    ys.append(y)

            if indices is None:
                x = np.vstack(xs)
                prev_y = np.vstack(prev_ys)
                y = np.vstack(ys)
            else:
                x = np.vstack(xs)[indices]
                prev_y = np.vstack(prev_ys)[indices]
                y = np.vstack(ys)[indices]

        if shuffle:
            raise NotImplementedError

        if isinstance(data, pd.DataFrame) and 'index' in data:
            data.pop('index')

        if self.ts_args['forecast_len'] == 1 and len(self.output_features)>0:
            y = y.reshape(-1, len(self.output_features))

        return x, prev_y, y

    def deindexify_nparray(self, data, key):
        if data.ndim == 3:
            _data, index = data[..., 1:].astype(np.float32), data[:, -1, 0]
        elif data.ndim == 2:
            _data, index = data[..., 1:].astype(np.float32), data[:, 0]
        elif data.ndim == 4:
            _data, index = data[..., 1:].astype(np.float32), data[:, -1, -1, 0]
        elif data.ndim == 5:
            _data, index = data[..., 1:].astype(np.float32), data[:, -1, -1, -1, 0]
        else:
            raise NotImplementedError

        if self.index_types[key] != 'str':
            index = np.array(index, dtype=np.int64)
        return _data, index

    def total_exs(self, 
                lookback,
                forecast_step=0, forecast_len=1, 
                **ts_args
                ):

        intervals = self.intervals
        input_steps = self.ts_args['input_steps']

        data = consider_intervals(self.data, intervals)

        num_outs = len(self.output_features) if self.output_features is not None else None

        max_tot_obs = 0
        if not self.allow_nan_labels and intervals is None:
            _data = data[self.input_features + self.output_features] if isinstance(data, pd.DataFrame) else data
            x, _, _ = prepare_data(_data,
                                lookback, num_outputs=num_outs,
                                forecast_step=forecast_step,
                                forecast_len=forecast_len, mask=np.nan, **ts_args)
            max_tot_obs = len(x)

        # we need to ignore some values at the start
        more = (lookback * input_steps) - 1

        if isinstance(data, np.ndarray):
            return len(data) - more

        # todo, why not when allow_nan_labels>0?
        if forecast_step > 0:
            more += forecast_step

        if forecast_len > 1:
            more += forecast_len

        if intervals is None: intervals = [()]

        more *= len(intervals)

        if self.allow_nan_labels == 2:
            tot_obs = data.shape[0] - more

        elif self.allow_nan_labels == 1:
            label_y = data[self.output_features].values
            idx = ~np.array([all([np.isnan(x) for x in label_y[i]]) for i in range(len(label_y))])
            tot_obs = np.sum(idx) - more
        else:

            if num_outs == 1:
                tot_obs = data.shape[0] - int(data[self.output_features].isna().sum()) - more
                tot_obs = max(tot_obs, max_tot_obs)

            else:
                # count by droping all the rows when nans occur in output features
                tot_obs = len(data.dropna(subset=self.output_features))
                tot_obs -= more

        return tot_obs

    def KFold_splits(self, n_splits=5):
        """returns an iterator for kfold cross validation.

        The iterator yields two tuples of training and test x,y pairs.
        The iterator on every iteration returns following
        `(train_x, train_y), (test_x, test_y)`
        Note: only `training_data` and `validation_data` are used to make kfolds.

        Example:
            >>> data = pd.DataFrame(np.random.randint(0, 10, (20, 3)), columns=['a', 'b', 'c'])
            >>> data_set = DataSet(data=data, config={'lookback': 1})
            >>> kfold_splits = data_set.KFold_splits()
            >>> for (train_x, train_y), (test_x, test_y) in kfold_splits:
            ...     print(train_x, train_y, test_x, test_y)

        """
        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")
        x, _, y = self._training_data()

        kf = KFold(n_splits=n_splits, 
                  random_state=self.seed if self.shuffle else None,
                  shuffle=self.shuffle)

        spliter = kf.split(x)

        for tr_idx, test_idx in spliter:

            yield (x[tr_idx], y[tr_idx]), (x[test_idx], y[test_idx])

    def LeaveOneOut_splits(self):
        """Yields leave one out splits
        The iterator on every iteration returns following
        `(train_x, train_y), (test_x, test_y)`"""
        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")
        x, _, y = self._training_data()

        kf = LeaveOneOut()

        for tr_idx, test_idx in kf.split(x):

            yield (x[tr_idx], y[tr_idx]), (x[test_idx], y[test_idx])

    def TimeSeriesSplit_splits(self, n_splits=5, **kwargs):
        """returns an iterator for TimeSeriesSplit.
        The iterator on every iteration returns following
        `(train_x, train_y), (test_x, test_y)`
        """
        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")
        x, _, y = self._training_data()

        tscv = TimeSeriesSplit(n_splits=n_splits, **kwargs)

        for tr_idx, test_idx in tscv.split(x):

            yield (x[tr_idx], y[tr_idx]), (x[test_idx], y[test_idx])

    def plot_KFold_splits(self, n_splits=5, show=True, **kwargs):
        """Plots the indices of kfold splits"""
        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")
        x, _, y = self._training_data()

        kf = KFold(n_splits=n_splits, 
                  random_state=self.seed if self.shuffle else None,
                  shuffle=self.shuffle)

        spliter = kf.split(x)

        self._plot_splits(spliter, x, title="KFoldCV", show=show, **kwargs)

        return

    def plot_LeaveOneOut_splits(self, show=True, **kwargs):
        """Plots the indices obtained from LeaveOneOut strategy"""
        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")
        x, _, y = self._training_data()

        spliter = LeaveOneOut().split(x)

        self._plot_splits(spliter=spliter,
                          x=x,
                          title="LeaveOneOutCV",
                          show=show,
                          **kwargs)

        return

    def plot_TimeSeriesSplit_splits(self, n_splits=5, show=True, **kwargs):
        """Plots the indices obtained from TimeSeriesSplit strategy"""

        if self.teacher_forcing:
            warnings.warn("Ignoring prev_y")

        x, _, y = self._training_data()

        spliter = TimeSeriesSplit(n_splits=n_splits, **kwargs).split(x)

        self._plot_splits(spliter=spliter,
                          x=x,
                          title="TimeSeriesCV",
                          show=show,
                          **kwargs)
        return

    def _plot_splits(self, spliter, x, show=True,  **kwargs):

        splits = list(spliter)

        figsize = kwargs.get('figsize', (10, 8))
        legend_fs = kwargs.get('legend_fs', 20)
        legend_pos = kwargs.get('legend_pos', (1.02, 0.8))
        title = kwargs.get("title", "CV")

        plt.close('all')
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

        for ii, split in enumerate(splits):
            indices = np.array([np.nan] * len(x))
            indices[split[0]] = 1
            indices[split[1]] = 0

            ax.scatter(range(len(indices)), [ii + .5] * len(indices),
                       c=indices, marker='_', lw=10, cmap="coolwarm",
                       vmin=-.2, vmax=1.2)

        yticklabels = list(range(len(splits)))

        ax.set(yticks=np.arange(len(splits)) + .5, yticklabels=yticklabels,
               xlabel='Sample index',
               ylabel="CV iteration")
        ax.set_title(title, fontsize=20)

        ax.legend([Patch(color=cmap_cv(.8)), Patch(color=cmap_cv(.02))],
                  ['Training set', 'Test set'],
                  loc=legend_pos, fontsize=legend_fs)

        if show:
            plt.show()
        return

    def to_disk(self, path: str = None):
        import h5py

        path = path or os.getcwd()
        filepath = os.path.join(path, "data.h5")

        f = h5py.File(filepath, mode='w')

        for k,v in self.init_paras().items():
            if isinstance(v, (dict, list, tuple, float, int, str)):
                f.attrs[k] = json.dumps(
                    v, default=jsonize).encode('utf8')

            elif v is not None and k != 'data':
                f.attrs[k] = v

        if self.teacher_forcing:
            x, prev_y, y = self.training_data()
            val_x, val_prev_y, val_y = self.validation_data()
            test_x, test_prev_y, test_y = self.test_data()
        else:
            prev_y, val_prev_y, test_prev_y = np.empty(0), np.empty(0), np.empty(0)
            x, y = self.training_data()
            val_x, val_y = self.validation_data()
            test_x, test_y = self.test_data()

        # save in disk
        self._save_data_to_hdf5('training_data', x, prev_y, y, f)

        self._save_data_to_hdf5('validation_data', val_x, val_prev_y, val_y, f)

        self._save_data_to_hdf5('test_data', test_x, test_prev_y, test_y, f)

        f.close()
        return

    def _save_data_to_hdf5(self, data_type, x, prev_y, y, f):
        """Saves one data_type in h5py. data_type is string indicating whether
        it is training, validation or test data."""

        assert x is not None
        group_name = f.create_group(data_type)

        container = {}
        container['x'] = x

        if self.teacher_forcing:
            container['prev_y'] = prev_y

        container['y'] = y

        for name, val in container.items():

            param_dset = group_name.create_dataset(name, val.shape, dtype=val.dtype)
            if not val.shape:
                # scalar
                param_dset[()] = val
            else:
                param_dset[:] = val
        return

    @classmethod
    def from_h5(cls, path):
        """Creates an instance of DataSet from .h5 file."""
        import h5py

        f = h5py.File(path, mode='r')

        config = {}
        for k, v in f.attrs.items():
            if isinstance(v, str) or isinstance(v, bytes):
                v = decode(v)
            config[k] = v

        cls._from_h5 = True
        f.close()

        # the data is already being loaded from h5 file so no need to save it again
        # upon initialization of class
        config['save'] = False
        return cls(path, **config)



#     The code in this file is modified after
#     https://github.com/dvolgyes/zenodo_get/blob/master/zenodo_get/zget.py

import sys
import os
import json
import signal
import time
import hashlib
from contextlib import contextmanager

try:
    import requests
except ModuleNotFoundError:
    requests = None

from .utils import download

abort_signal = False
abort_counter = 0
exceptions = False

def ctrl_c(func):

    signal.signal(signal.SIGINT, func)
    return func

@ctrl_c
def handle_ctrl_c(*args, **kwargs):
    global abort_signal
    global abort_counter
    global exceptions

    abort_signal = True
    abort_counter += 1

    if abort_counter >= 2:
        if exceptions:
            raise Exception('\n Immediate abort. There might be unfinished files.')
        else:
            sys.exit(1)


#see https://stackoverflow.com/questions/431684/how-do-i-change-the-working-directory-in-python/24176022#24176022
@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def check_hash(filename, checksum):
    algorithm, value = checksum.split(':')
    if not os.path.exists(filename):
        return value, 'invalid'
    h = hashlib.new(algorithm)
    with open(filename, 'rb') as f:
        while True:
            data = f.read(4096)
            if not data:
                break
            h.update(data)
    digest = h.hexdigest()
    return value, digest


def download_from_zenodo(outdir,
                         doi,
                         cont=False,
                         error=False,
                         **kwargs):
    """
    to suit the requirements of this package.
    :param outdir: Output directory, created if necessary. Default: current directory.
    :param doi: str, Zenodo DOI
    :param cont: True, Do not continue previous download attempt. (Default: continue.)
    :param error: False, Continue with next file if error happens.
    :param kwargs:
        sandbox: bool, Use Zenodo Sandbox URL.
        timeout: int, Connection time-out. Default: 15 [sec].
        pause: float, Seconds to wait before retry attempt, e.g. 0.5
        retry: int, Number of times to Retry on error.
    """
    _wget = kwargs.get('wget', None)
    md5 = kwargs.get('md5', False)
    keep = kwargs.get('keep', False)
    timeout = kwargs.get('timeout', 15)
    sandbox = kwargs.get('sandbox', False)
    pause = kwargs.get('pause', 0.5)
    retry = kwargs.get('retry', 0)

    with cd(outdir):

        url = doi
        if not url.startswith('http'):
            url = 'https://doi.org/' + url
        try:
            r = requests.get(url, timeout=timeout)
        except requests.exceptions.ConnectTimeout:
            raise TimeoutError("Connection timeout.")
        except Exception:
            raise ConnectionError
        if not r.ok:
            raise ValueError(f'DOI {doi} could not be resolved. Try again, or use record ID.')

        recordID = r.url.split('/')[-1]

        if not sandbox:
            url = 'https://zenodo.org/api/records/'
        else:
            url = 'https://sandbox.zenodo.org/api/records/'

        try:
            r = requests.get(url + recordID, timeout=timeout)
        except requests.exceptions.ConnectTimeout:
            raise TimeoutError('Connection timeout during metadata reading.')
        except Exception:
            raise ConnectionError('Connection error during metadata reading.')

        if r.ok:
            js = json.loads(r.text)
            files = js['files']
            total_size = sum(f['size'] for f in files)

            if md5 is not None:
                with open('md5sums.txt', 'wt') as md5file:
                    for f in files:
                        fname = f['key']
                        checksum = f['checksum'].split(':')[-1]
                        md5file.write(f'{checksum}  {fname}\n')

            if _wget is not None:
                if _wget == '-':
                    for f in files:
                        link = f['links']['self']
                        print(link)
                else:
                    with open(_wget, 'wt') as wgetfile:
                        for f in files:
                            fname = f['key']
                            link = 'https://zenodo.org/record/{}/files/{}'.format(
                                recordID, fname
                            )
                            wgetfile.write(link + '\n')
            else:
                print('Title: {}'.format(js['metadata']['title']))
                print('Keywords: ' +
                       (', '.join(js['metadata'].get('keywords', []))))
                print('Publication date: ' + js['metadata']['publication_date'])
                print('DOI: ' + js['metadata']['doi'])
                print('Total size: {:.1f} MB'.format(total_size / 2 ** 20))

                for f in files:
                    if abort_signal:
                        print('Download aborted with CTRL+C.')
                        print('Already successfully downloaded files are kept.')
                        break
                    link = f['links']['self']
                    size = f['size'] / 2 ** 20
                    print()
                    print(f'Link: {link}   size: {size:.1f} MB')
                    fname = f['key']
                    checksum = f['checksum']

                    remote_hash, local_hash = check_hash(fname, checksum)

                    if remote_hash == local_hash and cont:
                        print(f'{fname} is already downloaded correctly.')
                        continue

                    for _ in range(retry + 1):
                        try:
                            filename = download(link)
                        except Exception:
                            print('  Download error.')
                            time.sleep(pause)
                        else:
                            break
                    else:
                        print('  Too many errors.')
                        if not error:
                            raise Exception('Download is aborted. Too  many errors')
                        print('  Download continues with the next file.')
                        continue

                    h1, h2 = check_hash(filename, checksum)
                    if h1 == h2:
                        print(f'Checksum is correct. ({h1})')
                    else:
                        print(f'Checksum is INCORRECT!({h1} got:{h2})')
                        if not keep:
                            print('  File is deleted.')
                            os.remove(filename)
                        else:
                            print('  File is NOT deleted!')
                        if not error:
                            sys.exit(1)
                else:
                    print('All files have been downloaded.')
        else:
            raise Exception('Record could not get accessed.')



from ._main import Experiments
from ._rgr import MLRegressionExperiments
from ._cls import MLClassificationExperiments
from ._main import TransformationExperiments
from ._dl import DLRegressionExperiments



import warnings
from typing import Union

import numpy as np

try:
    from hyperopt import space_eval, hp
except ImportError:
    space_eval, hp = None, None


try:
    from skopt.space import Real as _Real
    from skopt.space import Integer as _Integer
    from skopt.space import Categorical as _Categorical
    import skopt
except ImportError:
    skopt, _Integer, _Categorical, _Real = None, None, None, None


try:
    from optuna.distributions import CategoricalDistribution, UniformDistribution, IntLogUniformDistribution
    from optuna.distributions import IntUniformDistribution, LogUniformDistribution
except ImportError:
    CategoricalDistribution, UniformDistribution, IntLogUniformDistribution = None, None, None
    IntUniformDistribution, LogUniformDistribution = None, None


from ai4water.utils.utils import Jsonize

# helper class to be able to print [1, ..., 4] instead of [1, '...', 4]
class _Ellipsis:
    def __repr__(self):
        return '...'


class Dummy:
    def __init__(self, low=None, high=None, *args, **kwargs):
        self.low = low
        self.high = high


if _Real is None:
    class _Real(Dummy): pass
    class _Integer(Dummy): pass
    class _Categorical(Dummy): pass


class Real(_Real):
    """
    This class is used for the parameters which have fractional values
    such as real values from 1.0 to 3.5.
    This class extends the `Real` class of Skopt so that it has an attribute grid which then
    can be fed to optimization algorithm to create grid space. It also adds several further
    methods to it.

    Attributes
    ------------
    grid

    Methods
    ----------
    - as_hp
    - to_optuna
    - suggest
    - to_optuna
    - serialize

    Example
    -------
        >>> from ai4water.hyperopt import Real
        >>> lr = Real(low=0.0005, high=0.01, prior='log', name='lr')
    """
    counter = 0
    def __init__(self,
                 low: float = None,
                 high: float = None,
                 num_samples: int = None,
                 step: int = None,
                 grid: Union[list, np.ndarray] = None,
                 *args,
                 **kwargs
                 ):
        """
        Arguments:
            low : lower limit of parameter
            high : upper limit of parameter
            step : used to define `grid` in conjuction with `low` and `high`
            grid : array like, if given, `low`, `high`, `step` and `num_samples`
                will be redundant.
            num_samples : if given, it will be used to create grid space using the formula
        """
        if low is None:
            assert grid is not None
            assert hasattr(grid, '__len__')
            low = np.min(grid)
            high = np.max(grid)

        self.counter += 1
        if 'name' not in kwargs:
            kwargs['name'] = f'real_{self.counter}'

        if skopt is not None:
            kwargs = check_prior(kwargs)

        self.num_samples = num_samples
        self.step = step
        super().__init__(low=low, high=high, *args, **kwargs)
        self.grid = grid

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, x):
        if x is None:
            if self.num_samples:
                self._grid = np.linspace(self.low, self.high, self.num_samples)
            elif self.step:
                self._grid = np.arange(self.low, self.high, self.step)
            else:
                self._grid = None
        else:
            self._grid = np.array(x)

    def as_hp(self, as_named_args=True):

        if self.prior == 'log-uniform':
            return hp.loguniform(self.name, low=self.low, high=self.high)
        else:
            assert self.prior in ['uniform', 'loguniform', 'normal', 'lognormal',
                                  'quniform', 'qloguniform', 'qnormal', 'qlognormal']
            if as_named_args:
                return getattr(hp, self.prior)(label=self.name, low=self.low, high=self.high)
            else:
                return getattr(hp, self.prior)(self.name, self.low, self.high)

    def suggest(self, _trial):
        # creates optuna trial
        log = False
        if self.prior:
            if self.prior == 'log':
                log = True

        return _trial.suggest_float(name=self.name,
                                    low=self.low,
                                    high=self.high,
                                    step=self.step,  # default step is None
                                    log=log)

    def to_optuna(self):
        """returns an equivalent optuna space"""
        if self.prior != 'log':
            return UniformDistribution(low=self.low, high=self.high)
        else:
            return LogUniformDistribution(low=self.low, high=self.high)

    def serialize(self):
        """Serializes the `Real` object so that it can be saved in json"""
        _raum = {k: Jsonize(v)() for k, v in self.__dict__.items() if not callable(v)}
        _raum.update({'type': 'Real'})
        return _raum

    def __repr__(self):
        return f"Real(low={self.low}, high={self.high}," \
               f" prior='{self.prior}', transform='{self.transform_}' name='{self.name}')"


class Integer(_Integer):
    """
    This class is used when the parameter is integer such as integer values from 1 to 10.
    Extends the Real class of Skopt so that it has an attribute grid which then
    can be fed to optimization algorithm to create grid space. Moreover it also
    generates optuna and hyperopt compatible/equivalent instances.

    Attributes
    ------------
    grid

    Methods
    ----------
    - as_hp
    - to_optuna
    - suggest
    - to_optuna
    - serialize

    Example:
        >>> from ai4water.hyperopt import Integer
        >>> units = Integer(low=16, high=128, name='units')

    """
    counter = 0
    def __init__(self,
                 low: int = None,
                 high: int = None,
                 num_samples: int = None,
                 step: int = None,
                 grid: "np.ndarray, list" = None,
                 *args,
                 **kwargs
                 ):
        """
        Arguments:
            low : lower limit of parameter
            high : upper limit of parameter
            grid list/array: If given, `low` and `high` should not be given as they will be
                calculated from this grid.
            step int: if given , it will be used to calculated grid using the formula
                np.arange(low, high, step)
            num_samples int: if given, it will be used to create grid space using the formula
                np.linspace(low, high, num_samples)
        """

        if low is None:
            assert grid is not None
            assert hasattr(grid, '__len__')
            low = np.min(grid)
            high = np.max(grid)

        self.counter += 1
        if 'name' not in kwargs:
            kwargs['name'] = f'integer_{self.counter}'

        self.num_samples = num_samples
        self.step = step

        if skopt is not None:
            kwargs = check_prior(kwargs)

        super().__init__(low=low, high=high, *args, **kwargs)
        self.grid = grid

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, x):
        if x is None:
            if self.num_samples:
                __grid = np.linspace(self.low, self.high, self.num_samples, dtype=np.int32)
                self._grid = [int(val) for val in __grid]
            elif self.step:
                __grid = np.arange(self.low, self.high, self.step, dtype=np.int32)
                self._grid = [int(val) for val in __grid]
            else:
                self._grid = None
        else:
            assert hasattr(x, '__len__'), f"unacceptable type of grid {x.__class__.__name__}"
            self._grid = np.array(x)

    def as_hp(self, as_named_args=True):
        if as_named_args:
            return hp.randint(self.name, low=self.low, high=self.high)
        else:
            return hp.randint(self.name, self.low, self.high)

    def suggest(self, _trial):
        # creates optuna trial
        log = False
        if self.prior:
            if self.prior == 'log':
                log = True

        return _trial.suggest_int(name=self.name,
                                  low=self.low,
                                  high=self.high,
                                  step=self.step if self.step else 1,  # default step is 1
                                  log=log)

    def to_optuna(self):
        """returns an equivalent optuna space"""
        if self.prior != 'log':
            return IntUniformDistribution(low=self.low, high=self.high)
        else:
            return IntLogUniformDistribution(low=self.low, high=self.high)

    def serialize(self):
        """Serializes the `Integer` object so that it can be saved in json"""
        _raum = {k: Jsonize(v)() for k, v in self.__dict__.items() if not callable(v)}
        _raum.update({'type': 'Integer'})
        return _raum

    def __repr__(self):
        return f"Integer(low={self.low}, high={self.high}," \
               f" prior='{self.prior}', transform='{self.transform_}' name='{self.name}')"


class Categorical(_Categorical):
    """
    This class is used when parameter has distinct group/class of values such
    as [1,2,3] or ['a', 'b', 'c']. This class overrides skopt's `Categorical` class.
    It Can be converted to optuna's distribution or hyper_opt's choice. It uses
    same input arguments as received by skopt's `Categorical`_ class

    Methods
    ----------
    - as_hp
    - to_optuna
    - suggest
    - to_optuna
    - serialize

    Example
    -------
        >>> from ai4water.hyperopt import Categorical
        >>> activations = Categorical(categories=['relu', 'tanh', 'sigmoid'], name='activations')


    .. _Categorical:
        https://scikit-optimize.github.io/stable/modules/generated/skopt.space.space.Categorical.html
    """
    @property
    def grid(self):
        return self.categories

    def as_hp(self, as_named_args=True):
        categories = self.categories
        if isinstance(categories, tuple):
            categories = list(self.categories)
        return hp.choice(self.name, categories)

    def suggest(self, _trial):
        # creates optuna trial
        return _trial.suggest_categorical(name=self.name, choices=self.categories)

    def to_optuna(self):
        return CategoricalDistribution(choices=self.categories)

    def serialize(self):
        """Serializes the `Categorical object` so that it can be saved in json"""
        _raum = {k: Jsonize(v)() for k, v in self.__dict__.items() if not callable(v)}
        _raum.update({'type': 'Integer'})
        return _raum

    def __repr__(self):
        if len(self.categories) > 7:
            cats = self.categories[:3] + (_Ellipsis(),) + self.categories[-3:]
        else:
            cats = self.categories

        if self.prior is not None and len(self.prior) > 7:
            prior = self.prior[:3] + [_Ellipsis()] + self.prior[-3:]
        else:
            prior = self.prior

        return f"Categorical(categories={cats}, prior={prior} name='{self.name}')"


def check_prior(kwargs: dict):
    prior = kwargs.get('prior', 'uniform')
    if prior in ["log"] and skopt.__version__ in ["0.9.0"]:
        # todo see why this will give error
        kwargs['prior'] = "log-uniform"
    return kwargs





__all__ = ['DataSetPipeline']

import os

import numpy as np

from ._main import _DataSet


class DataSetPipeline(_DataSet):
    """A collection of DataSets concatenated one after the other. A DataSetPipeLine
    of four DataSets will be as follows:

        +----------+
        | DataSet1 |
        +----------+
        | DataSet2 |
        +----------+
        | DataSet3 |
        +----------+
        | DataSet4 |
        +----------+

    The only condition for different datasets is that they have the same output dimension.

    """
    def __init__(
            self,
            *datasets,
            verbosity=1
    ) -> None:
        """
        Parameters
        ----------
        *datasets :
            the datasets to be combined
        verbosity :
            controls the output information being printed.

        """
        self.verbosity = verbosity

        self._datasets = []
        for ds in datasets:
            assert isinstance(ds, _DataSet), f"""
                            {ds} is not a valid dataset"""
            self._datasets.append(ds)

        self.examples = {}

        _DataSet.__init__(self, config={}, path=os.getcwd())

    @property
    def num_datasets(self) -> int:
        return len(self._datasets)

    @property
    def teacher_forcing(self):
        return all([ds.teacher_forcing for ds in self._datasets])

    @property
    def mode(self):
        return all([ds.mode for ds in self._datasets])

    @property
    def is_binary(self):
        return all([ds.is_binary for ds in self._datasets])

    @property
    def input_features(self):
        _input_features = [ds.input_features for ds in self._datasets]
        return _input_features

    @property
    def output_features(self):
        _output_features = [ds.output_features for ds in self._datasets]
        return _output_features

    def training_data(self, key="train", **kwargs):
        if self.teacher_forcing:
            x, prev_y, y = self._get_x_yy('training_data')
            return x, prev_y, y
        else:
            x, y = self._get_xy('training_data')
            return self.return_xy(np.row_stack(x), np.row_stack(y), "Training")

    def validation_data(self, key="val", **kwargs):
        if self.teacher_forcing:
            x, prev_y, y = self._get_x_yy('validation_data')
            return x, prev_y, y
        else:
            x, y = self._get_xy('validation_data')
            return self.return_xy(x, y, "Validation")

    def test_data(self, key="test", **kwargs):
        if self.teacher_forcing:

            x, prev_y, y = self._get_x_yy('test_data')
            return x, prev_y, y
        else:

            x, y = self._get_xy('test_data')
            return self.return_xy(x, y, "Test")

    def _get_x_yy(self, method):
        x, prev_y, y = [], [], []
        exs = {}

        for idx, ds in enumerate(self._datasets):
            _x, _prev_y, _y = getattr(ds, method)()
            x.append(_x)
            prev_y.append(_prev_y)
            y.append(_y)
            exs[idx] = {'x': len(x), 'y': len(y)}

        self.examples[method] = exs

        if not all([i.size for i in x]):
            x = conform_shape(x)
            prev_y = conform_shape(prev_y)
            y = conform_shape(y)

        return np.row_stack(x), np.row_stack(prev_y), np.row_stack(y)

    def _get_xy(self, method):

        x, y = [], []
        exs = {}
        for idx, ds in enumerate(self._datasets):
            _x, _y = getattr(ds, method)()
            x.append(_x)
            y.append(_y)
            exs[idx] = {'x': len(x), 'y': len(y)}

        self.examples[method] = exs

        if not all([i.size for i in x]):
            x = conform_shape(x)
            y = conform_shape(y)

        return np.row_stack(x), np.row_stack(y)


def conform_shape(alist:list):
    desired_shape = list([i.shape for i in alist if i.size != 0][0])
    desired_shape[0] = 0
    return [np.zeros(desired_shape) if arr.size == 0 else arr for arr in alist]



import tensorflow as tf
import tensorflow.keras.backend as K
import numpy as np


def reset_graph(seed=313):
    tf.compat.v1.reset_default_graph()
    tf.compat.v1.set_random_seed(seed)  # tf.random.set_seed(seed)  #
    np.random.seed(seed)


def tf_nse(true, _pred, name='NSE'):
    """ Nash-Sutcliff efficiency to be used as loss function. It is subtracted from one before being returned"""
    neum = tf.reduce_sum(tf.square(tf.subtract(_pred, true)))
    denom = tf.reduce_sum(tf.square(tf.subtract(true, tf.math.reduce_mean(true))))
    const = tf.constant(1.0, dtype=tf.float32)
    _nse = tf.subtract(const, tf.math.divide(neum, denom), name=name)
    return tf.subtract(const, _nse, name=name + '_LOSS')


def corr_coeff(true, predicted):
    """ Pearson correlation coefficient

    https://stackoverflow.com/a/58890795/5982232
    """
    mx = tf.math.reduce_mean(true)
    my = tf.math.reduce_mean(predicted)
    xm, ym = true - mx, predicted - my
    r_num = tf.math.reduce_mean(tf.multiply(xm, ym))
    r_den = tf.math.reduce_std(xm) * tf.math.reduce_std(ym)
    return r_num / r_den


def tf_kge(true, predicted):
    """ Kling Gupta efficiency. It is not being subtracted from 1.0 so that it can be used as loss"""
    tf_cc = corr_coeff(true, predicted)
    tf_alpha = tf.math.reduce_std(predicted) / tf.math.reduce_std(true)
    tf_beta = K.sum(predicted) / K.sum(true)
    return K.sqrt(K.square(tf_cc - 1.0) + K.square(tf_alpha - 1.0) + K.square(tf_beta - 1.0))


def tf_r2(true, predicted):
    """
    https://www.kaggle.com/c/mercedes-benz-greener-manufacturing/discussion/34019
    it is like r2_square score of sklearn which can be negative
    Not being subtracted from 1.0
    """
    r = corr_coeff(true, predicted)
    return r ** 2


def tf_r2_mod(true, predicted):
    """
    https://www.kaggle.com/c/mercedes-benz-greener-manufacturing/discussion/34019
    it is like r2_square score of sklearn which can be negative
    Not being subtracted from 1.0
    """
    ss_res = K.sum(K.square(true - predicted))
    ss_tot = K.sum(K.square(true - K.mean(true)))
    return ss_res / (ss_tot + K.epsilon())


def tf_nse_beta(true, predicted, name='nse_beta'):
    """
    Beta decomposition of NSE. See Gupta et. al 2009
    used in kratzert et al., 2018
    """
    const = tf.constant(1.0, dtype=tf.float32)
    nse_beta = (K.mean(predicted) - K.mean(true)) / K.std(true)
    return tf.subtract(const, nse_beta, name=name + '_LOSS')


def tf_nse_alpha(true, predicted, name='nse_alpha'):
    """
    Alpha decomposition of NSE. See Gupta et. al 2009
    used in kratzert et al., 2018
    It is being subtracted from 1.0
    """
    const = tf.constant(1.0, dtype=tf.float32)
    nse_alpha = K.std(predicted) / K.std(true)
    return tf.subtract(const, nse_alpha, name=name + '_LOSS')

def pbias(true, predicted):

    _sum = K.sum(tf.subtract(predicted, true))
    _a = tf.divide(_sum, K.sum(true))
    return 100.0 * _a


def nse(true, _pred, name='NSE'):
    """Nash-Sutcliff efficiency to be used as loss function. It is subtracted from one before being returned"""
    neum = tf.reduce_sum(tf.square(tf.subtract(_pred, true)))
    denom = tf.reduce_sum(tf.square(tf.subtract(true, tf.math.reduce_mean(true))))
    const = tf.constant(1.0, dtype=tf.float32)
    _nse = tf.subtract(const, tf.math.divide(neum, denom), name=name)
    return 1.0 - tf.subtract(const, _nse, name=name + '_LOSS')


def kge(true, predicted):
    """ Kling Gupta efficiency. It is not being subtracted from 1.0 so that it can be used as loss"""
    tf_cc = corr_coeff(true, predicted)
    tf_alpha = tf.math.reduce_std(predicted) / tf.math.reduce_std(true)
    tf_beta = K.sum(predicted) / K.sum(true)
    return 1.0 - K.sqrt(K.square(tf_cc - 1.0) + K.square(tf_alpha - 1.0) + K.square(tf_beta - 1.0))


# extracts features of 1d array like data.

import numpy as np
import scipy
from scipy.stats import norm, rankdata

class Features(object):

    def __init__(self, x):
        self.x = x

class Trends(Features):

    """
    Arguments:
    x array/list/series: 1d array or array like whose features are to be calculated.
    """

    def sen_slope(self, alpha=None):
        # https://github.com/USGS-python/trend/blob/master/trend/__init__.py
        """A nonparametric estimate of trend.
            Parameters
        ----------
        x : array_like
            Observations taken at a fixed frequency.
        Notes
        -----
        This method works with missing or censored data, as long as less <20% of
        observations are censored.
        References
        ----------
        .. [1] Helsel and Hirsch, R.M. 2002. Statistical Methods in Water Resources.
        .. [2] https://vsp.pnnl.gov/help/vsample/nonparametric_estimate_of_trend.htm
        """
        s = sen_diff(self.x)
        s.sort()

        if alpha:
            N = len(s)
            # calculate confidence limits
            C_alpha = norm.ppf(1 - alpha / 2) * np.sqrt(np.nanvar(self.x))
            U = int(np.round(1 + (N + C_alpha) / 2))
            L = int(np.round((N - C_alpha) / 2))
            return np.nanmedian(s), s[L], s[U]

        else:
            return np.nanmedian(s)

    def seasonal_sen_slope(self, period=12, alpha=None):
        """A nonparametric estimate of trend for seasonal time series.
        Paramters
        ---------
        x : array_like
            Observations taken at a fixed frequency.
        period : int
            Number of observations in a cycle. The number of seasons.
        """
        s = 0

        for season in np.arange(0, period):
            x_season = self.x[season::period]
            s = np.append(s, sen_diff(x_season))

        s.sort()

        if alpha:
            #  XXX This code needs to be verified
            N = len(s)
            # calculate confidence limits
            C_alpha = norm.ppf(1-alpha/2)*np.sqrt(np.nanvar(self.x))
            U = int(np.round(1 + (N + C_alpha)/2))
            L = int(np.round((N - C_alpha)/2))
            return np.nanmedian(s), s[L], s[U]

        else:
            return np.nanmedian(s)

    def pettitt(self, alpha=0.05):
        """Pettitt's change-point test
        A nonparameteric test for detecting change points in a time series.
        Parameters
        ----------
        x : array_like
            Observations taken at a fixed frequency.
        alpha : float
            Significance level
        Return
        ------
        The index of the change point of the series, provided that it is
        statistically significant.
        """
        U_t = np.zeros_like(self.x)
        n = len(self.x)

        r = rankdata(self.x)
        for i in np.arange(n):
            U_t[i] = 2 * np.sum(r[:i+1]) - (i+1)*(n-1)

        t = np.argmax(np.abs(U_t))
        K_t = U_t[t]

        p = 2.0 * np.exp((-6.0 * K_t**2)/(n**3 + n**2))

        if p > alpha:
            return t
        else:
            return np.nan

    def mann_kendall(self, alpha=0.05):
        """Mann-Kendall (MK) is a nonparametric test for monotonic trend.
        Parameters
        ----------
        x : array
            Observations taken at a fixed frequency.
        Returns
        -------
        z : float
            Normalized MK test statistic.
        Examples
        --------
        >>> x = np.random.rand(100) + np.linspace(0,.5,100)
        >>> z,p = kendall(x)
        Attribution
        -----------
        Modified from code by Michael Schramn available at
        https://github.com/mps9506/Mann-Kendall-Trend/blob/master/mk_test.py
        """
        # n = len(self.x)

        s = mk_score(self.x)
        var_s = mk_score_variance(self.x)

        z = mk_z(s, var_s)
        # calculate the p_value
        p_value = 2*(1-norm.cdf(abs(z)))  # two tail test

        return p_value


    def seasonal_mann_kendall(self, period=12):
        """ Seasonal nonparametric test for detecting a monotonic trend.
        Parameters
        ----------
        x : array
            A sequence of chronologically ordered observations with fixed
            frequency.
        period : int
            The number of observations that define period. This is the number of seasons.
        """
        # Compute the SK statistic, S, for each season
        s = 0
        var_s = 0

        for season in np.arange(period):
            x_season = self.x[season::period]
            s += mk_score(x_season)
            var_s += mk_score_variance(x_season)

        # Compute the SK test statistic, Z, for each season.
        z = mk_z(s, var_s)

        # calculate the p_value
        p_value = 2*(1-norm.cdf(abs(z)))  # two tail test

        return p_value


def mk_z(s, var_s):
    """Computes the MK test statistic, Z.
    Parameters
    ----------
    s : float
        The MK trend statistic, S.
    var_s : float
        Variance of S.
    Returns
    -------
    MK test statistic, Z.
    """
    # calculate the MK test statistic
    if s > 0:
        z = (s - 1)/np.sqrt(var_s)
    elif s < 0:
        z = (s + 1)/np.sqrt(var_s)
    else:
        z = 0

    return z

def mk_score_variance(x):
    """Computes corrected variance of S statistic used in Mann-Kendall tests.
    Equation 8.4 from Helsel and Hirsch (2002).
    Parameters
    ----------
    x : array_like
    Returns
    -------
    Variance of S statistic
    Note that this might be equivalent to:
        See page 728 of Hirsch and Slack
    References
    ----------
    .. [1] Helsel and Hirsch, R.M. 2002. Statistical Methods in Water Resources.
    """
    x = x[~np.isnan(x)]
    n = len(x)
    # calculate the unique data
    unique_x = np.unique(x)
    # calculate the number of tied groups
    g = len(unique_x)

    # calculate the var(s)
    if n == g:  # there is no tie
        var_s = (n * (n - 1) * (2 * n + 5)) / 18

    else:  # there are some ties in data
        tp = np.zeros_like(unique_x)
        for i in range(len(unique_x)):
            tp[i] = sum(x == unique_x[i])
        var_s = (n * (n - 1) * (2 * n + 5) - np.sum(tp * (tp - 1) * (2 * tp + 5))) / 18

    return var_s


class Stats(Features):

    def auc(self):
        return

    def auto_corr(self):
        return

    def centroid(self):
        return

    def slope(self):
        return

    def zero_crossing_rate(self):
        return

    def sum_abs_diff(self):
        return np.sum(np.abs(np.diff(self.x)))

    def min_max_diff(self):
        return np.abs(np.max(self.x) - np.min(self.x))

    def mean_abs_dev(self):
        return np.mean(np.abs(self.x - np.mean(self.x, axis=0)), axis=0)

    def median_abs_dev(self):
        """Median absolute deviation"""
        return scipy.stats.median_absolute_deviation(self.x, scale=1)

    def rms(self):
        """Root mean square"""
        return np.sqrt(np.sum(np.array(self.x) ** 2) / len(self.x))

def mk_score(x):
    """Computes S statistic used in Mann-Kendall tests.
    Parameters
    ----------
    x : array_like
        Chronologically ordered array of observations.
    Returns
    -------
    MK trend statistic (S).
    """
    x = x[~np.isnan(x)]
    n = len(x)
    s = 0

    for j in np.arange(1, n):
        s += np.sum(np.sign(x[j] - x[0:j]))

    return s

def sen_diff(x):
    """Sen's difference operator.
    Paramaters
    ----------
    x : array_like
        Observations taken at a fixed frequency.
    Returns
    -------
    Sen difference
    """
    #x = x[~np.isnan(x)]
    n = len(x)
    N = int(n*(n-1)/2)  # number of slope estimates
    s = np.zeros(N)
    i = 0
    for j in np.arange(1, n):
        #s[i:j+i] = (x[j] - x[0:j])/np.arange(1, j+1)
        s[i:j+i] = (x[j] - x[0:j])/np.arange(j, 0, -1)
        i += j

    return s


if __name__ == "__main__":
    f = Features(np.random.random(10))
